/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact.  */

/* Interface definitions for curses xconq. */

/* Termcap defn "SE" conflicts with direction "SE". */

#undef SE

#include <curses.h>

#define VIEWFILE "view.cconq"

enum inputmode {
    SURVEY = 0,
    MOVE = 1,
    HELP = 2,
    MORE = 3,
    PROMPT = 4
  };

enum screenmode {
    normalwins = 0,
    helpwin = 1,
    sidecloseupwin = 2
  };

enum listtype {
    byrandorder = 0,
    bytype = 1,
    byactorder = 2,
    bylocation = 3,
    numlisttypes = 4
  };

enum listsides {
    ourside = 0,
    ourallies= 1,
    allsides = 2,
    numlistsides = 3
  };

#define INFOLINES 4

/* The user interface substructure.  This is only allocated for sides with
   attached displays. */

typedef struct a_ui {

    short followaction;       /* move to where a change has occured */

    int gotkey;

    int reqx, reqy;
    int requtype;

    int curx, cury;
    int tmpcurx, tmpcury;

    Unit *curunit;
    Unit *tmpcurunit;

    enum inputmode mode;
    enum inputmode prevmode;

    enum screenmode screenmode;

    int itertime;

    char inpch;               /* keyboard char */

    /* Random input data slots - flush these! */
    struct a_side *reqside;
    struct a_unit *requnit;
    int reqvalue, reqvalue2;
    char *reqdeflt;           /* A default string */
    int reqposx, reqposy;     /* accumulator for position reading */

    int cmdarg;

    char ustr[MAXUTYPES];     /* used in composing unit type hints */
    int uvec[MAXUTYPES];      /* vector of allowed unit types to input */
    int bvec[MAXUTYPES];      /* bit vector of allowed unit types to input */

    /* Constructed during display init */
    short nw, nh;             /* length and number of notice lines */
    /* Working variables for the display */
    short lastx, lasty;       /* last current x and y (-1,-1 initially) */

    char text1[BUFSIZE];      /* current prompt on display */
    char text2[BUFSIZE];      /* current prompt on display */
    int reqstrbeg;
    int reqstrend;            /* data about string under construction */

    int ouremblem;          /* pattern for national emblem (pixmap?) */
    int emblems[MAXSIDES];

    struct a_helpnode *curhelpnode;
    int help;                /* help window */
    char *helpstring;
    int helpscreentop;

    int wx[20], wy[20], ww[20], wh[20];
    int nextwin;                /* number of next window to open */

    int date;
    int sides;                /* overall game info */
    int toplines;                 /* places for notices/warnings/prompts */
    int clock;               /* chess clock display */
    int map;                 /* Display of a piece of the world */
    int list;                /* Unit listing */
    int closeup;
    int sidecloseup;

    /* Map dimensions. */
    short mw, mh;
    short totw, toth;
    short sx, sy;
    short vx, vy;             /* lower left of viewport in hex coords */
    short vw, vh;             /* Boundaries of viewport in hex coords */
    short vw2, vh2;           /* 1/2 (rounded down) of above values */
    short vcx, vcy;           /* center of the view */
    short lastvcx, lastvcy;
    /* List dimensions. */
    short lw, lh;
    /* Side list dimensions. */
    short sh;
    /* How to draw the map. */
    int drawterrain;         /* Display solid color terrain on the map? */
    int drawunits;           /* Display units on the map? */
    int drawnames;           /* Display unit names/numbers on the map? */
    int drawpeople;
    int drawelevations;
    int drawweather;

    enum listtype listtype;               /* what this list is displaying. */
    enum listsides listsides;
    int test;
    int value;
    int sorton;
    int sortorder;

    struct a_unit *unit;

    struct a_side *side;

    int active;

#ifdef DESIGNERS
    short curutype;
    short curttype;
    short curside;
    short curfeature;
#endif
} UI;

/* Magic values (unlikely to show up as player input) */

#define DFLTFLAG (-12345)       /* flags when default arg to be used */
#define NEGFLAG  (-12346)       /* flags when default arg to be used */
#define ITERFLAG (-12347)       /* flags args that are "itertime" */

#define ESCAPE '\033'           /* standard abort character */
#define BACKSPACE '\010'        /* for fixing strings being entered */

extern Side *dside;

extern UI dui;

extern int usebothchars;

#define append_to_buffer(x,y) strcat(x,y)

#ifdef DEBUGGING
#define UNSEEN_CHAR '?'
#else
#define UNSEEN_CHAR ' '
#endif

#define terrain_visible(x, y)  \
  (g_see_all() || dside->designer || terrain_view(dside, wrapx(x), y) != UNSEEN)

#define units_visible(x, y)  \
  (g_see_all() || dside->designer || cover(dside, wrapx(x), y) >= 1)
