/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

#include "conq.h"
#include "cconq.h"

int usebothchars = TRUE;

/* Completely redo a screen, making no assumptions about appearance.
   This is a last-gasp measure, most redrawing should be restricted
   to only the directly affected windows.  Also this shouldn't be
   done without the user's permission, since it will blow away impending
   input. */

redraw()
{
    if (active_display(dside)) {
	switch (dui.screenmode) {
	  case normalwins:
	    clear();
	    draw_divider();
	    show_toplines();
	    show_game_date();
	    show_side_list();
	    show_list();
	    show_closeup();
	    show_map();
	    /* (should flush any typeahead here?) */
	    break;
	  case helpwin:
	    show_help();
	    break;
	  case sidecloseupwin:
	    break;
	  default:
	    break;
	}
	show_cursor();
	refresh();
    }
}

draw_divider()
{
    int i;

    for (i = 2; i < LINES; ++i) {
	mvaddstr(i, dui.mw, "|");
    }
}

/* Decide whether given location is not too close to edge of screen. */
/* We do this because it's a pain to move units when half the adjacent */
/* places aren't even visible.  This routine effectively places a lower */
/* limit of 5x5 for the map window. (I think) */

in_middle(x, y)
int x, y;
{
    int sx, sy;

    xform(x, y, &sx, &sy);
    return (between (3, sx, dui.mw - 3) && between (3, sy, dui.mh - 3));
}

xform(x, y, sxp, syp)
int x, y, *sxp, *syp;
{
    if (in_area(x, y)) {
	*sxp = x * 2 + y - dui.sx;
	*syp = (dui.toth - y) - dui.sy;
    } else {
	/* Don't complain, just make screwups obvious on the screen. */
	*sxp = *syp = 1;
    }
}

/* Display a map and all of its paraphernalia. */

show_map()
{
    int y1, y2, y, x1, x2, adj;
    int halfheight = area.height / 2;

    clear_window(dui.map);
    set_map_viewport();
    /* Compute top and bottom rows to be displayed. */
    y1 = min(dui.vy + dui.vh, area.height - 1);
    y2 = dui.vy;
    for (y = y1; y >= y2; --y) {
	/* Adjust the right and left bounds to fill the viewport as
	   much as possible, without going too far (the drawing code
	   will clip, but clipped drawing is still expensive). */
	x1 = dui.vx - (y - dui.vy) / 2;
	x2 = x1 + dui.vw + 2;
	/* If the area doesn't wrap, then we might have to stop
	   drawing before we reach the edge of the viewport. */
	if (!area.xwrap) {
	    x1 = max(0, min(x1, area.width - 1));
	    x2 = max(0, min(x2, area.width));
	    if (x1 + y > area.width + halfheight)
	      continue;
	    if (x2 + y < halfheight)
	      continue;
	    if (x2 + y > area.width + halfheight)
	      x2 = area.width + halfheight - y;
	    if (x1 + y < halfheight)
	      x1 = halfheight - y;
	}
	draw_row(x1, y, x2 - x1);
    }
    /* Draw modeline in standout if possible. */
    standout();
    sprintf(spbuf, "--Map--%s-----------------------------------------",
	    dui.mode == SURVEY ? "Survey" : "Move");
    draw_text(dui.map, 0, dui.mh - 1, spbuf);
    standend();
}

draw_row(x0, y0, len)
int x0, y0, len;
{
    int x;

    if (dui.drawterrain) {
	draw_terrain_row(x0, y0, len);
    }
    /* Draw sparse things on top of the basic row. */
    for (x = x0; x < x0 + len; ++x) {
	/* The relative ordering of these is quite important.  Note that
	   each should be prepared to run independently also, since the
	   other displays might have been turned off. */
	if (people_sides_defined() && dui.drawpeople) {
	    draw_people(x, y0);
	}
	if (dui.drawunits) {
	    draw_units(x, y0);
	}
    }
    /* Need a second loop so names are superimposed properly. */
    for (x = x0; x < x0 + len; ++x) {
	if (dui.drawnames) {
	    draw_legend(x, y0);
	}
    }
}

/* Draw a single row of just terrain. */

draw_terrain_row(x0, y0, len)
int x0, y0, len;
{
    char ch, ch2;
    int x, sx, sy, t, t2;

    xform(x0, y0, &sx, &sy);
    for (x = x0; x < x0 + len; ++x) {
	ch = ((x % 2 == 0 && y0 % 2 == 0) ? UNSEEN_CHAR : ' ');
	ch2 = ' ';
	if (terrain_visible(x, y0)) {
	    t = terrain_at(x, y0);
	    ch = ((t_char(t) != NULL && strlen(t_char(t)) > 0)
		  ? t_char(t)[0]
		  : t_type_name(t)[0]);
	    ch2 = (usebothchars ? ch : ' ');
	    /* Just provide a hint at presence of aux terrain. */
	    if (any_aux_terrain_defined()) {
		for_all_terrain_types(t2) {
		    if (aux_terrain_defined(t2)) {
			if (aux_terrain_at(x, y0, t2)) {
			    ch2 = '#';
			    break;
			}
		    }
		}
	    }
	}
	if (cur_at(dui.map, sx, sy)) {
	    addch(ch);
	}
	if (cur_at(dui.map, sx + 1, sy)) {
	    addch(ch2);
	}
	sx += 2;
    }
}

/* Draw a single unit char pair as appropriate. */

draw_units(x, y)
int x, y;
{
    int draw = FALSE;
    char ch;
    int sx, sy, uview, u, s;
    Unit *unit;
    
    if (g_see_all()) {
	if ((unit = unit_at(x, y)) != NULL) {
	    u = unit->type;
	    s = side_number(unit->side);
	    draw = TRUE;
	}
    } else if ((uview = unit_view(dside, wrap(x), y)) != EMPTY) {
	u = vtype(uview);
	s = vside(uview);
	draw = TRUE;
    }
    if (draw) {
	xform(x, y, &sx, &sy);
	if (cur_at(dui.map, sx, sy)) {
	    ch = ((u_uchar(u) != NULL && strlen(u_uchar(u)) > 0)
		  ? u_uchar(u)[0]
		  : utype_name_n(u, 1)[0]);
	    addch(ch);
	    if (between(1, s, numsides)) addch(s + '0');
	}
    }
}

/* Indicate what kind of people are living in the given hex. */

draw_people(x, y)
int x, y;
{
    int pop, sx, sy;
    Side *side2;

    if (terrain_visible(x, y)
	&& (pop = people_side_at(wrapx(x), y)) != NOBODY) {
	xform(x, y, &sx, &sy);
	cur_at(dui.map, sx + 1, sy);
	addch(pop + '0');
    }
}

/* Draw any text that should be associated with this hex. */

/* (could precompute what the string will lap over and move or truncate str),
   should be deterministic for each mag, so redraw doesn't scramble */

/* do geofeatures, label at a hex with nothing else, and declared as the
   feature's "center" */

draw_legend(x, y)
int x, y;
{
    int xx, numhexes = 1, color, sx, sy;
    char legend[100];
    int uview = unit_view(dside, x, y);
    Unit *unit;

    if (uview == EMPTY)  return;
    /* Draw a unit's name or number. */
    /* (This will only do top unit in hex, what about others?) */
    if (dui.drawunits
	&& (unit = unit_at(x, y)) != NULL) {
	/* still has a bug if types happen to match - need to use view date
	   instead? */
	if (vtype(uview) != unit->type) return;
	if (unit->name != NULL) {
	    strcpy(legend, unit->name);
	} else if (unit->number > 0) {
	    sprintf(legend, "%d%s",
		    unit->number, ordinal_suffix(unit->number));
	} else {
	    /* Turns out we have nothing to make a legend with. */
	    return;
	}
	/* Look at all the hexes that the legend will overlap. */
	numhexes = strlen(legend) / 2;
	xform(x, y, &sx, &sy);
	draw_text(dui.map, sx + 2, sy, legend);
    } else {
	/* feature object should specify legend's position */
	extern GeoFeature *feature_at();
	GeoFeature *feature;

	feature = feature_at(x, y);
	if (feature) {
	    if (feature->size == 1) {
		int pixlen;

		strcpy(legend, feature_name_at(x, y));
		pixlen = strlen(legend) / 2;
		xform(x, y, &sx, &sy);
		draw_text(dui.map, sx + 1, sy, legend);
	    }
	}
    }
}

/* (should blink the cursor or something to indicate a hit) */

draw_blast(unit, es, hit)
Unit *unit;
Side *es;
int hit;
{
#if 0
    int ux = unit->x, uy = unit->y, sx, sy, i, btype;
    Side *us = unit->side, *side;
    int map;

    /* "if area_hit"? */
    if (0 /* hit >= period_nukehit() */) {
#if 0
	/* or all sides? */
	if (active_display(us)) invert_whole_map(us);
	if (active_display(es)) invert_whole_map(es);
	sleep(1);
	/* may need a time delay if X is too speedy */
	if (active_display(us)) invert_whole_map(us);
	if (active_display(es)) invert_whole_map(es);
	for (i = 0; i < 4; ++i) {
	    if (active_display(us)) draw_mushroom(us, ux, uy, i);
	    if (active_display(es)) draw_mushroom(es, ux, uy, i);
	    if (i != 2 && (active_display(us) || active_display(es))) sleep(1);
	}
	if (active_display(us) || active_display(es)) sleep(1);
#endif
    } else {
	btype = ((hit >= unit->hp) ? 2 : ((hit > 0) ? 1 : 0));
	for_all_sides(side) {
	    if (active_display(side) && cover(side, ux, uy) > 0) {
		    if (4 >= 4) { /* what other tests? */
			xform(ux, uy, &sx, &sy);
			/* should be able to draw different size blasts
			   for each map zoom */
#if 0
			draw_blast_icon(dui.map,
					sx + (2 / 2 - 10),
					sy + (1 / 2 - 11),
					btype);
#endif
			refresh();
		    } else {
			/* just flicker the screen a little? */
		    }
	    }
	}
    }
#endif
}

/* A predicate that tests whether our display can safely be written to. */
/* (should this maybe be a macro?) */

active_display(side)
Side *side;
{
    return (side == dside && side_has_display(side) && dui.active);
}

/* General window clearing. */

clear_window(win)
int win;
{
    int i;

    if (dui.wx[win] == 0
	&& dui.wy[win] == 0
	&& dui.ww[win] == COLS
	&& dui.wh[win] == LINES) {
	clear();
    } else if (between(0,dui.wx[win],COLS-1)
	       && between(0,dui.ww[win],COLS-1)
	       && between(0,dui.wy[win],LINES-1)
	       && between(0,dui.wh[win],LINES-1)) {
	for (i = 0; i < dui.ww[win]; ++i) tmpbuf[i] = ' ';
	tmpbuf[dui.ww[win]] = '\0';
	for (i = 0; i < dui.wh[win]; ++i)
	  mvaddstr(dui.wy[win]+i, dui.wx[win], tmpbuf);
    } else {
	printf("error: win %d is %dx%d @ %d,%d\n",
	       win, dui.ww[win], dui.wh[win],
	       dui.wx[win], dui.wy[win]);
    }
}

/* Draw a large blot over the area. */

draw_mushroom(x, y, i)
int x, y, i;
{
    int sx, sy;

    xform(x, y, &sx, &sy);
    if (cur_at(dui.map, sx, sy)) {
	addstr("##");
	refresh();
	if (i > 0) {
	    if (cur_at(dui.map, sx-1, sy+1)) addstr("####");
	    if (cur_at(dui.map, sx-2, sy)) addstr("######");
	    if (cur_at(dui.map, sx-1, sy-1)) addstr("####");
	    refresh();
	}
    }
}

/* Drawing text is easy, but we do need to do clipping manually. */

draw_text(win, x, y, str)
int win;
int x, y;
char *str;
{
    int i, linestart = 0;

    if (y < 0) {
	y = dui.wh[win] - y;
    }
    if (cur_at(win, x, y)) {
	for (i = 0; i < strlen(str); ++i) {
	    if (str[i] == '\n') {
		if (y < dui.wh[win]) {
		    cur_at(win, x, ++y);
		    linestart = i;
		} else {
		    /* Ran out of room */
		    return i;
		}
	    } else if (x + (i - linestart) < dui.ww[win]) {
		addch(str[i]);
	    }
	}
    }
    return (-1);
}

/* Make a beep by writing ^G. */

beep()
{
    putchar('\007');
}

/* Shut a single display down - presumably only called once. */

close_display(side)
Side *side;
{
    clear();
    refresh();
    endwin();
    dui.active = FALSE;
    DGprintf("Display \"%s\" closed.\n", side->player->displayname);
}

/* Put the current view into a file. */

/* (should be intelligent enough to cut into pages, or else document
   how to do it) */
/* (maybe display names too somehow, perhaps as second layer??) */

dump_view()
{
#if 0
    char ch1, ch2;
    int x, y, i, vs;
    viewdata view;
    Side *side2;
    FILE *fp;

    if ((fp = fopen(VIEWFILE, "w")) != NULL) {
	for (y = area.height-1; y >= 0; --y) {
	    for (i = 0; i < y; ++i) fputc(' ', fp);
	    for (x = 0; x < area.width; ++x) {
		view = side_view(x, y);
		if (view == UNSEEN) {
		    ch1 = ch2 = ' ';
		} else if (view == EMPTY) {
		    ch1 = t_char(terrain_at(x, y));
		    ch2 = ch1; /* but not always maybe? */
		} else {
		    ch1 = '*'; /* unit char */
		    vs = vside(view);
		    side2 = side_n(vside(view));
		    ch2 = (side2 ? ((side == side2) ? ' ' : vs + '0') : '`');
		}
		fputc(ch1, fp);
		fputc(ch2, fp);
	    }
	    fprintf(fp, "\n");
	}
	fclose(fp);
	notify("Dumped area view to \"%s\".", VIEWFILE);
    } else {
	notify("Can't open \"%s\"!!", VIEWFILE);
    }
#endif
}

show_help()
{
    int i, x = 0, y = 0, linestart = 0, win = dui.help;
    char *str = dui.helpstring;

    clear();
    if (cur_at(win, x, y)) {
	for (i = 0; i < strlen(str); ++i) {
	    if (str[i] == '\n') {
		if (y < dui.wh[win]) {
		    cur_at(win, x, ++y);
		    linestart = i;
		} else {
		    cur_at(win, x, y-2);
		    standout();
		    addstr("--More--");
		    standend();
		    /* and record spot? */
		}
	    } else if (x + (i - linestart) < dui.ww[win]) {
		addch(str[i]);
	    }
	}
	cur_at(win, 0, dui.wh[win] - 1);
	addstr("['n' for next topic, 'p' for previous, 'q' to end]");
    }
}

/* Put the terminal's cursor at an appropriate place. */

show_cursor()
{
    int sx, sy;

    if (active_display(dside)) {
	switch (dui.mode) {
	  case SURVEY:
	  case MOVE:
	    xform(dui.curx, dui.cury, &sx, &sy);
	    cur_at(dui.map, sx, sy);
	    break;
	  case HELP:
	    cur_at(dui.help, 0, 0);
	    break;
	  case MORE:
	    beep();
	    break;
	  case PROMPT:
	    /* This doesn't account for two-line prompts. */
	    cur_at(dui.toplines, strlen(dui.text1), 0);
	    break;
	  default:
	    cur_at(dui.toplines, 0, 0);
	}
	refresh();
    }
}

/* Don't just position the cursor, but also clip and return the decision. */

cur_at(win, x, y)
int win, x, y;
{
    int sx, sy;

    if (x < 0 || x >= dui.ww[win] || y < 0 || y >= dui.wh[win]) {
	move(0, 0);
	return FALSE;
    } else {
	sy = dui.wy[win] + y;  sx = dui.wx[win] + x; 
	if (between(0, sx, COLS-1) && between(0, sy, LINES-1)) {
	    move(sy, sx);
	} else {
	    printf("error: win %d at %d,%d -> %d,%d\n", win, x, y, sx, sy);
	}
	return TRUE;
    }
}
