/* Copyright (c) 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Definitions for the Mac interface. */

#undef USE_CMDLINE

#ifdef THINK_C
#include <MacHeaders>
#endif /* THINK_C */

#ifndef THINK_C /* assume MPW */
#include <Values.h>
#include <Types.h>
#include <Resources.h>
#include <QuickDraw.h>
#include <Fonts.h>
#include <Events.h>
#include <Windows.h>
#include <Menus.h>
#include <TextEdit.h>
#include <Dialogs.h>
#include <Desk.h>
#include <ToolUtils.h>
#include <Memory.h>
#include <SegLoad.h>
#include <Files.h>
#include <Folders.h>
#include <OSUtils.h>
#include <OSEvents.h>
#include <DiskInit.h>
#include <Packages.h>
#include <Traps.h>
#include <Lists.h>
#include <PPCToolbox.h>
#endif

#ifndef THINK_C
#define QD(whatever) (qd.##whatever)
#else
#define QD(whatever) (whatever)
#endif

#define top_left(rect) (*(Point *) &(rect.top))
#define bottom_right(rect) (*(Point *) &(rect.bottom))

#include "macdefs.h"
#include "macimf.h"

/* How many magnifications we're supporting. */

#define NUMPOWERS 8

/* The kinds of sort keys available for list windows. */

enum listsortkeys {
	bynothing,
	bytype,
	byactorder,
	bylocation,
	byside
};

/* The types of available designer tools. */

enum tooltype {
	notool,
	terraintool,
	unittool,
	peopletool,
	materialtool,
	featuretool,
	elevationtool,
	temperaturetool,
	cloudstool,
	windstool
};

/* The user interface substructure.  This is only allocated for sides with
   attached displays. */

typedef struct a_ui {
	short followaction;	   /* move to where a change has occured */
	int itertime;
	int active;
} UI;

/* Each side can open up any number and shape of maps. */

#define MAXSELECTIONS 500

typedef struct a_map {
	short sx, sy;
	short totsw, totsh;
	short pxw, pxh;		    /* size of window in pixels */
	short vx, vy;			/* lower left of viewport in hex coords */
	short vw, vh;			/* Boundaries of viewport in hex coords */
	short power;			/* index to this map's magnification */
	short mag;				/* magnification of each hex (a power of 2) */
	short vcx, vcy;		    /* center of the view */
	short hw, hh;			/* pixel dims of a hex cell */
	short hch;				/* height of hex cell between centers */
	short uw, uh;			/* pixel dims of unit subcell */
	short toph;
	WindowPtr window;
	ControlHandle hscrollbar;
	ControlHandle vscrollbar;
	/* How to draw the map. */
	int drawterrain;		/* Display solid color terrain on the map? */
	int drawgrid;		    /* Draw outlines around hexes? */
	int drawhexpats;		/* Draw terrain patterns? */
	int drawothermaps;		/* Draw outlines of other maps */
	int drawunits;		    /* Display units on the map? */
	int drawnames;		    /* Display unit names/numbers on the map? */
	short oldesttoshow;	    /* the relative time of the oldest data */
	short agetofade;		/* point at which to gray out the display */
	short newesttoshow;	    /* the relative time of the newest data */
	short fadeterrain;		/* fade world features as well as units? */
	short lowestaltitude;   /* the lowest altitude of units to display */
	short highestaltitude;	/* the highest altitude of units to display */
	int drawborders;
	int drawconns;
	int drawpeople;
	int drawelevations;
	int drawmaterials[MAXMTYPES];
	int drawlighting;
	short drawtemp;
	short drawweather;
	int numselections;
	Unit *selections[500];
	Rect contentrect;
	int maxdepth;           /* Largest # of bits/pixel of screens this map uses */
	struct a_map *next;		/* Link to the next map. */
} Map;

/* Each side can open up any number of lists of units. */

#define MAXINLIST 500

typedef struct a_list {
#if (MAXSIDES < 32)
	long sides;                      /* mask of sides to display the units of */
#else
	short sides[MAXSIDES];
#endif
	enum listsortkeys sortkeys[5];  /* attributes to sort list elements on */
	int listglimpsed;              /* list glimpsed units also? */
	Unit *units[MAXINLIST];
	int selected[MAXINLIST];
	int numunits;
	int firstvisible;
	int lastvisible;
	int largeicons;
	int shouldreorg;
	WindowPtr window;
	ControlHandle hscrollbar;
	ControlHandle vscrollbar;
	struct a_list *next;
} List;

/* A closer look at a unit. */

typedef struct a_unit_closeup {
	struct a_unit *unit;
	WindowPtr window;
	struct a_unit_closeup *next;
} UnitCloseup;

/* A closer look at a side. */

typedef struct a_side_closeup {
	struct a_side *side;
	DialogPtr dialog;
	struct a_side_closeup *next;
} SideCloseup;

/* Iteration over all of a side's maps. */

#define for_all_maps(m)  \
  for ((m) = maplist; (m) != NULL; (m) = (m)->next)

/* Iteration over all of a side's lists. */

#define for_all_lists(l)  \
  for ((l) = listlist; (l) != NULL; (l) = (l)->next)

/* Iteration over all of a side's side closeups. */

#define for_all_side_closeups(c)  \
  for ((c) = sidecloseuplist; (c) != NULL; (c) = (c)->next)

#define for_all_unit_closeups(c)  \
  for ((c) = unitcloseuplist; (c) != NULL; (c) = (c)->next)

extern Side *dside;

extern int mags[], hws[], hhs[], hcs[], uws[], uhs[];
extern int dolin[];
extern int bsx[NUMPOWERS][7], bsy[NUMPOWERS][7], lsx[NUMPOWERS][6], lsy[NUMPOWERS][6];

Map *map_from_window();
List *list_from_window();
SideCloseup *side_closeup_from_window();

#define pascalify(STR) \
  sprintf(tmpbuf, " %s", STR);  \
  tmpbuf[0] = strlen(STR);

extern ImageFamily *uimages;
extern ImageFamily *timages;
extern ImageFamily *eimages;

extern ImageColor **tcolors;

extern DialogPtr helpwin;
extern struct a_helpnode *curhelpnode;
extern char *helpstring;
extern int helpscreentop;
extern TEHandle helptext;

extern WindowPtr gamewin;
extern ControlHandle gamevscrollbar;

extern WindowPtr buildwin;
extern ListHandle buildunitlist;
extern ListHandle buildtypelist;

extern WindowPtr designwin;
extern enum tooltype tooltype;
extern short curutype;
extern short curttype;
extern short cursidenumber;
extern short curfid;
extern GeoFeature *curfeature;

extern WindowPtr historywin;

extern MenuHandle sidemenu;
extern MenuHandle utypemenu;
extern MenuHandle mtypemenu;
extern MenuHandle ttypemenu;

extern int hasColorQD;

extern int maxscreendepth;

extern int sbarwid;

extern int bwid[];
extern int cwid[];

extern WindowPtr *winmenuwins;

extern int defaultdrawgrid;		   /* Display grid on the map? */
extern int defaultdrawnames;		   /* Display unit names/numbers on the map? */

extern int modaltool;

extern int didamove;

extern int moveonclick;

extern PolyHandle polygons[];
extern int lastpolyx[], lastpolyy[];
extern PolyHandle gridpolygons[];
extern int lastgridpolyx[], lastgridpolyy[];

extern struct a_map *maplist;	   /* chain of maps that we're using */
extern struct a_list *listlist;	 /* chain of lists */
extern struct a_unit_closeup *unitcloseuplist;	 /* chain of unit closeups */
extern struct a_side_closeup *sidecloseuplist;	 /* chain of side closeups */

extern CursHandle paintcursors[];

extern CursHandle cellpaintor;
extern CursHandle bordpaintor;
extern CursHandle connpaintor;
extern CursHandle unitpaintor;
extern CursHandle peoplepaintor;
extern CursHandle featurepaintor;
extern CursHandle materialpaintor;
extern CursHandle elevpaintor;
extern CursHandle cloudpaintor;
extern CursHandle temppaintor;

extern CursHandle movecursors[];
extern CursHandle opencrosscursor;

char *savefile_name();
char *newsfile_name();

UnitCloseup *unit_closeup_from_window();
UnitCloseup *find_unit_closeup();

