/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* This file manages the machine-run sides' overall behavior, that is to
   say, its strategy.  */

#include "conq.h"
#include "mplay.h"

extern int max_range;

/* Flag to detect when mplayer init has been done. */

int mplayerinited = FALSE;

mplayer_init(side)
Side *side;
{
	Unit *unit;

    if (!mplayerinited) {
	mplayer_init_shared();
	mplayerinited = TRUE;
	Dprintf("One mplayer AI is %d bytes.\n", sizeof(Strategy));
    }
    /* Make sure a strategy object exists. */
    if (mplayer(side) == NULL) create_strategy(side);
    if (!side_has_units(side)) return;
    analyze_the_game(side);
    /* Reset plans of any units that were not doing anything. */
    for_all_side_units(side, unit) {
    	if (in_play(unit) && unit->plan) {
    		unit->plan->asleep = FALSE;
    		unit->plan->waitingfortasks = FALSE;
    		if (unit->plan->type == FOLLOW_ORDERS
    			&& unit->plan->tasks == NULL) {
    			unit->plan->type = UNFORMED_PLAN;
    		}
    	}
    }
}

/* At the beginning of each turn, can make plans and review the situation.
   This should be frequent at first, but rather expensive to do always,
   so only some chance of doing it on a random turn. */

mplayer_init_turn(side)
Side *side;
{
    int x, y;

    /* Cases where we no longer need to run. */
    if (!side->ingame) return;
    if (!side_has_units(side)) return;
    /* Make sure a strategy object exists. */
    if (mplayer(side) == NULL) create_strategy(side);
    /* Look over the game design we're playing with. */
    analyze_the_game(side);
    DMprintf("%s mplayer init turn\n", side_desig(side));
    if (mplayer(side)->trytowin) {
	/* (should be able to create a new strategy periodically) */
	/* Check out the current goal tree first. */
	review_goals(side);
	/* Check out our exploration status. */
	review_search_zones(side);
	/* Check out all the theaters. */
	review_theaters(side);
	/* Check out all of our units. */
	review_units(side);
	/* Decide on the new current plan. */
	update_side_plan(side);
	/* Propagate this to individual unit plans. */
	update_unit_plans(side);
	/* Periodically check if our position is hopeless.  Do after planning,
	   since planning info might give us new optimism. */
	if (curturn > 20 || probability(25)) {
	    decide_resignation(side);
	}
    } else {
	update_unit_plans_randomly(side);
    }
    DMprintf("%s mplayer done with init turn\n", side_desig(side));
}

/* Create and install an entirely new strategy object for the side. */

create_strategy(side)
Side *side;
{
    Strategy *strategy = (Strategy *) xmalloc(sizeof(Strategy));

    /* Put the specific structure into a generic slot. */
    side->ai = (struct a_ai *) strategy;
/*    side->demand = (short *) xmalloc(numutypes * sizeof(short));  */
    strategy->type = mplayertype;
    /* Remember when we did this. */
    strategy->creationdate = curturn;
    /* Null out various stuff. */
    strategy->goalstates = NULL;
    strategy->searchzones = NULL;
    strategy->theaters = NULL;
    strategy->analyzegame = TRUE;
    /* Analyze the game and decide our basic goals. */
    analyze_the_game(side);
}

/* Look over the game design and decide what we're supposed to be doing,
   if anything at all.  This just sets up toplevel goals based on the
   game design, does not evaluate goals or any such. */

analyze_the_game(side)
Side *side;
{
    Side *side2;
    Goal *goal;
    GoalState *goalstate;

    if (mplayer(side)->analyzegame) {
	if (should_try_to_win(side)) {
	    mplayer(side)->trytowin = TRUE;
	    /* This is our whole purpose in the game. */
	    goal = create_goal(WON_GAME);
	    add_goalstate(side, create_goalstate(goal, side, 0));
	    /* (should add special goals for agreements here.) */
	    /* We want to "kick butt" - *everybody* else's butt. */
	    for_all_sides(side2) {
		if (side != side2) {
		    goal = create_goal(LOST_GAME);
		    add_goalstate(side, create_goalstate(goal, side2, 0));
		}
	    }
	    /* We might develop a sudden interest in exploration. */
	    if (!g_see_all()) {
		goal = create_goal(WORLD_KNOWN);
		add_goalstate(side, create_goalstate(goal, side, 0));
		for_all_sides(side2) {
		    if (side != side2) {
			goal = create_goal(POSITIONS_KNOWN);
			add_goalstate(side, create_goalstate(goal, side2, 0));
		    }
		}
		/* Also have the general goal of knowing where indeps are. */
		goal = create_goal(POSITIONS_KNOWN);
		add_goalstate(side, create_goalstate(goal, NULL, 0));
	    }
	} else {
	    mplayer(side)->trytowin = FALSE;
	}
	determine_subgoals(side);
	/* (should only be set false if machine thinks it can "win"?) */
	side->willingtodraw = FALSE;
	mplayer(side)->analyzegame = FALSE;
    }
}

determine_subgoals(side)
Side *side;
{
}

/* Find out how much remains to be explored in each search zone. */

review_search_zones(side)
Side *side;
{
    int i, x, y;
    short view;
    SearchZone *zone;

    if (!g_see_all() /* should check for goal instead */
	&& mplayer(side)->searchzones == NULL) {
	create_search_zones(side);
    }
    if (mplayer(side)->searchzones != NULL) {
	for_all_search_zones(side, i) {
	    search_zone_n(side, i, zone, x, y);
	    zone->unexplored = 0;
	}
	for_all_interior_hexes(x, y) {
	    if (inside_world(x, y)) {
		view = terrain_view(side, x, y);
		if (view == UNSEEN) {
		    ++(search_zone_at(side, x, y)->unexplored);
		}
		/* (should also considered if terrain seen but units not) */
	    }
	}
	for_all_search_zones(side, i) {
	    search_zone_n(side, i, zone, x, y);
	    if (zone->unexplored > (zone->size / 10)) {
		/* (should also guess at terrain amts here) */
		++(mplayer(side)->explorersneeded);
	    }
	    if (zone->unexplored != zone->size) {
		DMprintf("%s zone at %d %d: %d of %d cells unseen\n",
			 side_desig(side), x, y, zone->unexplored, zone->size);
	    }
	}
	DMprintf("%s other zones are entirely unknown\n", side_desig(side));
    }
}

create_search_zones(side)
Side *side;
{
    int i, x, y, u, t;
    Strategy *strategy = mplayer(side);
    SearchZone *zone;

    /* shape of zones should be adjusted to be similar to world shape? */
    strategy->zonewidth = 10;
    strategy->zoneheight = 10;
    /* Add 1 to count the partial zones around the edges. */
    strategy->numzonex =
      max(1, (area.width + strategy->zonewidth - 1) / strategy->zonewidth);
    strategy->numzoney =
      max(1, (area.height + strategy->zoneheight - 1) / strategy->zoneheight);
    strategy->numzones = strategy->numzonex * strategy->numzoney;
    /* Create and initialize the search zones. */
    strategy->searchzones = 
      (SearchZone *) xmalloc(strategy->numzones * sizeof(SearchZone));
    strategy->explorertypes = (short *) xmalloc(numutypes * sizeof(short));
    strategy->explorersneeded = 0;
    strategy->terrainguess = (short *) xmalloc(numttypes * sizeof(short));
    for_all_search_zones(side, i) {
	search_zone_n(side, i, zone, x, y);
	zone->size = 0;
	zone->unexplored = 0;
    }
    /* Compute how many valid hexes are in each zone. */
    for_all_interior_hexes(x, y) {
	if (inside_world(x, y)) {
	    ++(search_zone_at(side, x, y)->size);
	}
    }
    /* Reset the summation of our exploration needs. */
    for_all_unit_types(u) {
	strategy->explorertypes[u] = 0;
	strategy->explorersneeded = 0;
    }
    for_all_terrain_types(t) {
	strategy->terrainguess[t] = 0;
    }
}

/* Do a combination of analyzing existing theaters and creating new ones. */

/* (should theaters be more fluid if there are no sessile units?) */

review_theaters(side)
Side *side;
{
    int x, y, num, u, d, d2;
    short view;
    Unit *unit;
    Side *loop_side;
    Theater *theater;

    for_all_theaters(side, theater) {
	theater->allied_units = 0;
	theater->allied_makers = 0;
	theater->neutral_makers = 0;
	theater->makers = 0;
	theater->unexplored = 0;
	theater->border = FALSE;
	theater->allied_bases = 0;
	for_all_unit_types(u) {
	    theater->units[u] = NULL;
	    theater->requested_units[u] = 0;
	}
	theater->enemy_strength /= 2;
	theater->units_lost /= 2;
	theater->enemy_seen_recently = 0;
    }
    /* Now look at all the units that we can. */
    for_all_units (loop_side, unit) {
	if (allied_side(unit->side, side)) {
	    if ((theater = theater_at(side, x, y)) != NULL) {
		theater->allied_units++;
		theater->units[unit->type] = unit;
		if (0 /* u_maker(unit->type) */) theater->allied_makers++;
		if (isbase(unit)) theater->allied_bases++;
		if (0 /* u_maker(unit->type) */) {
		    theater->makers++;
/*		    if (indep(unit)) theater->neutral_makers++;  */
		}
	    }
	}
    }
    /* Now dig up what are only images. */
    for_all_interior_hexes(x, y) {
	if ((theater = theater_at(side, x, y)) != NULL) {
	    view = unit_view(side, x, y);
	    if (view == UNSEEN) {
		theater->unexplored++;
	    } else {
/*		if (view == EMPTY && side_view_age(side, x, y) < 4)
		  view = side_prevview(side, x, y);  */
		if (view != EMPTY && view != UNSEEN) {
		    u = vtype(view);
		    if (enemy_side(side_n(vside(view)), side)) {
			theater->enemy_seen_recently += unit_strength(u);
		    }
		}
	    }
	}
    }
    for_all_theaters(side, theater) {
	theater->enemy_strength += theater->enemy_seen_recently;
    }
    find_border_areas(side);
    /* if nothing unexplored in exploration theater, then it can be removed? */
    for_all_theaters(side, theater) {
	DMprintf("Theater \"%s\" at %d,%d (radius %d) unexplored %d\n",
		 theater->name, theater->x, theater->y, theater->radius[0],
		 theater->unexplored);
    }
}

#if 0
    /* Should this ever not exist? (if no sessile units?) */
    theater = create_theater(side);
    theater->name = "Home Front";
    theater->x = side->countryx;  theater->y = side->countryy;
    for_all_directions(d) {
	theater->radius[d] = g_min_radius();
    }
#endif

Theater *
create_theater(side)
Side *side;
{
    Theater *theater = (Theater *) xmalloc(sizeof(Theater));

    theater->next = mplayer(side)->theaters;
    mplayer(side)->theaters = theater;
    theater->name = "?";
/*    theater->purpose = HANG_OUT;  */
    theater->maingoalstate = NULL;
    return theater;
}

Theater *
theater_at(side, x, y)
Side *side;
int x, y;
{
    Theater *theater;

    for_all_theaters(side, theater) {
	if (distance(x, y, theater->x, theater->y) <= theater->radius[0]) {
	    return theater;
	}
    }
    return NULL;
}

/* Some notion of the unit's "strength"? */

unit_strength(ut)
int ut;
{
    return ((ave_build_time[ut]) ? ave_build_time[ut] : 10);
}

/* Identify those theaters that are at the edges of our territory. */

find_border_areas(side)
Side *side;
{
    int num, x, y, i;
    int safe;
    Theater *theater;

    for_all_theaters(side, theater) {
	if (theater->allied_bases == 0 ||
	    theater->unexplored > 0 ||
	    theater->allied_makers != theater->makers ||
	    theater->enemy_strength > 5) {
	    theater->border = TRUE;
	    theater->safe_area = FALSE;
	} else {
	    theater->nearby = TRUE;
	    theater->border = FALSE;
	    safe = theater->allied_units > 15;
#if 0
	    /* ? */
	    for_all_directions(i) {
		x = wrap(theater->x + dirx[i] * AREA_SIZE);
		y = interior(theater->y + diry[i] * AREA_SIZE);
		area_info[oparea_index(x,y)].nearby = TRUE;
		if (area_info[oparea_index(x,y)].allied_bases < 2) {
		    theater->border = TRUE;
		    safe = FALSE;
		}
	    }
	    mplayer(side)->areas[num].safe_area = safe;
#endif
	}
    }
}

/* Examine the goals to see what has been accomplished and what still needs
   to be done. */

review_goals(side)
Side *side;
{
    Strategy *strategy = mplayer(side);
    GoalState *goalstate;

    for_all_side_goalstates(side, goalstate) {
	/* (should be able to skip eval sometimes) */
	goalstate->state = goal_truth(side, goalstate->side, goalstate->goal);
	goalstate->evaldate = curturn;
    }
    /* Say something about each goal here. */
    for_all_side_goalstates(side, goalstate) {
	DMprintf("%s has %s\n",
		 side_desig(side), goalstate_desig(goalstate));
    }
    /* Should look at certainty of each goal and decide whether to keep or
       drop it, and mention in debug output also. */
}

/* Go through all our units (and allied ones). */

review_units(side)
Side *side;
{
    int u, i;
    Unit *unit;
    Side *loopside;
    
    side->numunits = 0;
    for_all_unit_types(u) {
	mplayer(side)->unitlist[u] = NULL;
	mplayer(side)->unitlistcount[u] = 0;
    }
    for_all_units(loopside, unit) {
	if (allied_side(unit->side, side)) {
	    side->numunits++;
	    u = unit->type;
	    mplayer(side)->unitlistcount[u]++;
	    /* chains ally units together?? */
/*	    unit->mlist = mplayer(side)->unitlist[u];  */
	    mplayer(side)->unitlist[u] = unit;
	}
    }
    /* Review individual unit goals. */
    for_all_side_units(side, unit) {
	if (unit->plan) {
	    Plan *plan = unit->plan;

	    /* Goal might have become satisfied. */
	    if (plan->maingoal) {
		if (goal_truth(side, side, plan->maingoal) == 100) {
		    DMprintf("%s %s satisfied, removing\n",
			     unit_desig(unit),
			     goal_desig(plan->maingoal));
		    plan->maingoal = NULL;
		}
	    }
	    /* Goal might have become satisfied. */
	    if (plan->maingoalstate) {
		if (goal_truth(side, side, plan->maingoalstate->goal) == 100) {
		    DMprintf("%s %s satisfied, removing\n",
			     unit_desig(unit),
			     goalstate_desig(plan->maingoalstate));
		    plan->maingoalstate = NULL;
		}
	    }
	}
    }
    /* Could notify display about unit mix? */
}

/* Look at our overall plan and hack it as needed. */

update_side_plan(side)
Side *side;
{
    Theater *theater;

    Dprintf("%s updating plan\n", side_desig(side));
    mplayer(side)->defend_priority = 50;
    /* only if world is unknown? */
    mplayer(side)->explore_priority = max(75 - curturn, 0);
    /* wait until we contact somebody? */
    mplayer(side)->attack_priority = max(0, 50 - mplayer(side)->explore_priority);
    if (mplayer(side)->searchzones != NULL) {
	decide_search_needs(side);
    }
    /* Add something to add/update theaters as things open up. */
    for_all_theaters(side, theater) {
	decide_theater_needs(side, theater);
    }
}

/* Look at the different search zones and decide what sort of units would
   be best suited to exploration. */

decide_search_needs(side)
Side *side;
{
    int i, x, y, u;
    SearchZone *zone;

    for_all_unit_types(u) {
	mplayer(side)->explorersneeded = 0;
    }
    if (1 /* need to continue search? */) {
	for_all_search_zones(side, i) {
	    search_zone_n(side, i, zone, x, y);
	    /* Look at zones, decide which utypes to go for. */
	    if (zone->unexplored > (zone->size / 10)) {
		++(mplayer(side)->explorersneeded);
	    }
	}
    }
    if (mplayer(side)->explorersneeded > 0) {
	DMprintf("%s needs %d explorers\n", side_desig(side),
		 mplayer(side)->explorersneeded);
    }
}

/* Figure out how many units to request for each area. */

decide_theater_needs(side, theater)
Side *side;
Theater *theater;
{
    if (theater->enemy_strength == 0
	&& theater->unexplored > 0) {
	/* Should look for good exploration units. */
	theater->importance = 50;  /* should depend on context */
/*	theater->reinforce = EXPLORE_AREA;  */
    } else if (theater->enemy_strength < 5) {
	if (theater->allied_makers == 0
	    && theater->makers > 0
	    && theater->nearby) {
	    theater->reinforce = GUARD_BORDER_TOWN + 2 * theater->makers;
	} else if (theater->makers > 0) {
	    theater->reinforce = (theater->border ? GUARD_BORDER_TOWN :
				  GUARD_TOWN) + 2 * theater->allied_makers;
	} else if (theater->allied_bases > 0) {
	    theater->reinforce = (theater->border ? GUARD_BORDER: GUARD_BASE);
	} else if (theater->border) {
	    theater->reinforce = NO_UNITS;
	} else {
	    theater->reinforce = NO_UNITS;
	}
    } else {
	if (theater->allied_makers > 0) {
	    theater->reinforce = DEFEND_TOWN + 5 * theater->makers;
	} else if (theater->allied_bases > 0) {
	    theater->reinforce = DEFEND_BASE + theater->allied_bases;
	} else {
	    theater->reinforce = 0 /* DEFEND_AREA */;
	}
    }
}

/* For each unit, decide what it should be doing (if anything).  This is
   when a side takes the initiative;  a unit can also request info from
   its side when it is working on its individual plan. */

update_unit_plans(side)
Side *side;
{
    Unit *unit;

    for_all_side_units(side, unit) {
	mplayer_decide_plan(side, unit);
    }
}

/* Randomly change a unit's plans. */

update_unit_plans_randomly(side)
Side *side;
{
    Unit *unit;

    for_all_side_units(side, unit) {
	if (unit->plan) {
	    if (probability(10)) {
		DMprintf("Randomly changed %s plan %s",
			 unit_desig(unit), plan_desig(unit->plan));
		unit->plan->type = random((int) NUMPLANTYPES);
		DMprintf("to plan %s\n", plan_desig(unit->plan));
	    } else if (probability(10)) {
		/* destroy and reconsitute plan? */
		/* alter random parts of a plan? */
	    }
	    if (probability(10)) {
		unit->plan->asleep = FALSE;
	    }
	}
    }
}

/* This makes a list of units friendly to the given side. */

/* It also calculates the units in each region. */

make_unitlists(side)
Side *side;
{
    Unit *unit;
    int ut, i;
    Side *loop_side;
    
    side->numunits = 0;
    for_all_unit_types(ut) {
	mplayer(side)->unitlist[ut] = NULL;
	mplayer(side)->unitlistcount[ut] = 0;
    }
#ifdef REGIONS      
    for_all_unit_types(ut) {
	for (i = 0; i < num_regions[ut]; i++)
	  units_in_region[ut][i] = 0;
    }
#endif
   for_all_units(loop_side, unit) {
       if (side == unit->side /* ? */) {
#if 0
	   if (!producing(unit) && probability(10)) {
	       wake_unit(unit, FALSE, WAKEOWNER, (Unit *) NULL);
	   }
#endif
       }
       if (allied_side(unit->side, side)) {
	   side->numunits++;
	   ut = unit->type;
	   mplayer(side)->unitlistcount[ut]++;
/*	   unit->mlist = mplayer(side)->unitlist[ut];  */
	   mplayer(side)->unitlist[ut] = unit;
#ifdef REGIONS
	   if (num_regions[ut] > 0) {
	       if (unit->side == side &&
		   (i = aref(unit_region[ut], unit->x, unit->y)) >= 0)
		 units_in_region[ut][i]++;
	   }
#endif
       }
   }
}

/* Send units to the different areas. */

assign_units(side)
Side *side;
{
#if 0
    int total_units, num, total_requests, i, u, priority, d;
    Unit *unit, *prev;

    if (!mplayer(side)) return;
    total_units = 0;
    for_all_side_units(side, unit) {
	if (mobile(unit->type)) {
	    total_units += unit_strength(unit->type);
	}
    }

    for (i = 0; i < NUMTOPAREAS; i++) {
	top_areas[i] = NULL;
    }
    total_requests = 0;
    for (num = 0; num < areas_wide * areas_high; num++) {
	total_requests += theater->reinforce;
    }

    if (total_requests > 0) {
	/* Figure out how many units to request of each type. */
	for (num = 0; num < areas_wide * areas_high; num++) {
	    for_all_unit_types(u) {
		if (mobile(u)) {
		    theater->requested_units[u] =
		      ((theater->reinforce) * (mplayer(side)->unitlistcount[u]) +
		       total_requests - total_requests/4) 
			/ total_requests;
		    theater->num_units[u] = 0;
		} else theater->requested_units[u] = 0;
	    }
	}
    }
    /* Allocate units to locations where they are already going if they */
    /* are really needed there.  If not, then allocate them to the */
    /* current location. */
    /* Figure out how many units are already allocated. */
    for_all_unit_types(u) {
	if (!mobile(u)) { /* No need to allocate. */
	    mplayer(side)->unitlist[u] = NULL;
	} else {
	    for (priority = 1; priority <= 2; priority++) {
		allocate_as_is(side, u, priority);
	    }
	    /* Mark all remaining units as unselected.  We will try to
	       assign them to the nearest and highest priority task */
/*	    for (unit = mplayer(side)->unitlist[u]; unit != NULL; unit = unit->mlist) {
		unit->plan->selection_status = 10000;
	    }  */
	    /* This loop looks for high priority units that are not
	       at their destinations.  All units being placed are guaranteed
	       not to be placed in the area they currently are in. */
	    for (num = 0; num < areas_wide * areas_high; num++)
	      if (needs_units(1, side, num, u) || needs_units(2, side, num, u)) {
		  prev = NULL;
		  /* Send unit to priority 2 locations only if the distance is */
		  /* half as far as the nearest priority 1 location. */
		  i = (needs_units(1, side, num, u)) ? 2 : 1;
		  for (unit = mplayer(side)->unitlist[u]; unit != NULL; unit = unit->mlist) {
		      d = distance(unit->x, unit->y, theater->x, theater->y) / i;
		      if (d < unit->plan->selection_status) {
			  if (unit->plan->selection_status < 10000) {
			      area_info[unit->plan->area].num_units[u]--;
			  }
			  unit->plan->area = num;
			  unit->plan->selection_status = d;
			  theater->num_units[u]++;
			  /* delete the unit immediately if it is close. no need */
			  /* to check any further */
			  if (d < AREA_SIZE) {
			      if (prev == NULL)
				mplayer(side)->unitlist[u] = unit->mlist;
			      else prev->mlist = unit->mlist;
			  } else prev = unit;
		      } else prev = unit;
		  }
	      }
	    for (unit = mplayer(side)->unitlist[u]; unit != NULL; unit = unit->mlist) {
		prev = NULL;
		if (unit->plan->selection_status < 10000) {
		    if (prev == NULL)
		      mplayer(side)->unitlist[u] = unit->mlist;
		    else prev->mlist = unit->mlist;
		} else prev = unit;
	    }
	    
	    /* Now allocate all other units alreay in the correct locations. */
	    allocate_as_is(side, u, 3);

	    for (num = 0; num < areas_wide * areas_high; num++)
	      if (needs_units(3, side, num, u)) {
		  prev = NULL;
		  for (unit = mplayer(side)->unitlist[u]; unit != NULL; unit = unit->mlist) {
		      d = distance(unit->x, unit->y, theater->x, theater->y);
		      if (d < unit->plan->selection_status) {
			  if (unit->plan->selection_status < 10000) {
			      area_info[unit->plan->area].num_units[u]--;
			  }
			  unit->plan->area = num;
			  unit->plan->selection_status = d;
			  theater->num_units[u]++;
			  if (d < AREA_SIZE) {
			      if (prev == NULL)
				mplayer(side)->unitlist[u] = unit->mlist;
			      else prev->mlist = unit->mlist;
			  } else prev = unit;
		      } else prev = unit;
		  }
	      }
#if 0
/* No needed, unit lists redone from start */
      for (unit = mplayer(side)->unitlist[u]; unit != NULL; unit = unit->mlist) {
	prev = NULL;
	if (unit->plan->selection_status < 10000) {
	  if (prev == NULL)
	    mplayer(side)->unitlist[u] = unit->mlist;
	  else prev->mlist = unit->mlist;
	} else prev = unit;
      }
#endif
	}
    }
    /* Allocate remaining units. */
    
    /* Unit allocation. */
    
    /* Allocate all units in war zones to that zone, unless they are */
    /* moving to another war zone.  

       Allocate units to other zones up to some fraction, provided that
       they have already been allocated to that area.
       
       Allocate the other units to the neediest area nearby.  War zones
       already have most of the units in them, so distance is not too
       important. 
       */
#endif
}

/* Are units for this area a top priority. */

needs_units(priority, side, area, utype)
int priority;
Side *side;
int area, utype;
{
#if 0
    switch (priority) {
    case 1:
	if (mplayer(side)->areas[area].enemy_strength < 4 ||
	    area_info[area].allied_bases == 0)
	  return FALSE;
	else if (area_info[area].requested_units[utype] >
		 area_info[area].num_units[utype])
	  return TRUE;
	else return FALSE;
	break;
      case 2:
	if (mplayer(side)->areas[area].enemy_strength < 4 ||
	    area_info[area].allied_makers == 0)
	  return FALSE;
	else if (area_info[area].requested_units[utype] / 2 >
		 area_info[area].num_units[utype])
	  return TRUE;
	else return FALSE;
	break;
      case 3: 
	if (area_info[area].requested_units[utype] >
	    area_info[area].num_units[utype])
	  return TRUE;
	else return FALSE;
	break;
      default:
	return TRUE;
	break;
    }
#endif
}

/* Change units' areas to ones that need them (more, or just need?) */

allocate_as_is(side, u, priority)
Side *side;
int u, priority;
{
#if 0
    Unit *unit, *prev;
    int num;

    prev = NULL;
    for (unit = mplayer(side)->unitlist[u]; unit != NULL; unit = unit->mlist) {
	num = oparea_index(unit->x, unit->y);
	if (unit->plan->area >= 0 && needs_units(priority, side, unit->plan->area, u)) {
	    area_info[unit->plan->area].num_units[u]++;
	    if (prev == NULL)
	      mplayer(side)->unitlist[u] = unit->mlist;
	    else
	      prev->mlist = unit->mlist;
	} else if (needs_units(priority, side, num, u)) {
	    unit->plan->area = num;
	    theater->num_units[u]++;
	    if (prev == NULL)
	      mplayer(side)->unitlist[u] = unit->mlist;
	    else
	      prev->mlist = unit->mlist;
	} else prev = unit;	  
    }
#endif
}

/* Returns true if an area has makers but nobody is scheduled to try to */
/* capture them. (what about assault?) */

area_needs_capturer(x, y)
int x, y;
{
#if 0
    int num = oparea_index(x, y);

    if (theater->makers < theater->allied_makers &&
	mmplayer(side) &&
	(mmplayer(side)->areas[num].capturers_approaching == 0 ||
	 mmplayer(side)->areas[num].capture_time - curturn > 10))
      return TRUE;
    else
#endif
      return FALSE;
}

/* Sometimes there is no point in going on, but be careful not to be too
   pessimistic.  Right now we only give up if no hope at all. */

/* (This needs a lot more work - because a submarine can sometimes hit
   a city, machine won't give up if only has subs left.) */

/* (This should be driven by particular scorekeepers) */

decide_resignation(side)
Side *side;
{
    int u, u2, couldwin = FALSE, coulddraw = FALSE, opposed, own, chance = 0;
    Side *side1, *side2;
    Strategy *strategy = mplayer(side);
    Unit *unit;

    /* For machine players only. */
    if (!mplayer(side)) return;
    /* Compute how our strengths compare. */
    make_estimates(side);
#if 0
    /* Now see if we still have a chance to win.  We (allies) need to
       have a unit that can make more units or a unit that can capture
       other units.  (This could be more precise.) */
    /* We can force a draw if we still have some units. */
    for_all_unit_types(u) {
	own = strategy->alstrengths[side_number(side)][u];
	/* This test should be a little tighter, so will resign if only
	   have submarines vs strong enemy or some such. */
	if (own > 0) {
	    coulddraw = TRUE;
	}
	for_all_unit_types(u2) {
	    /* should test if the type we make can beat the enemy types */
	    if (could_make(u, u2) && mobile(u2)) {
		couldwin = TRUE;
	    }
	    for_all_sides(side1) {
		if (enemy_side(side, side1)) {
		    opposed = strategy->alstrengths[side_number(side1)][u2];
		    if (own > 0 && opposed > 0) {
			if (could_capture(u, u2) && mobile(u))
			  couldwin = TRUE;
			if (could_hit(u, u2) && mobile(u))
			  couldwin = TRUE;
		    }
		}
	    }
	}
    }
#if 0
    /* Now decide what to do. */
    if (couldwin) {
	notify(side, "No need to resign");
	/* We're still in the running. */
    } else if (coulddraw) {
	notify(side, "Could end with a draw");
	/* maybe think about giving up? */
    } else {
	notify(side, "Time to give up!");
	give_up(side);
    }
#endif
#endif
}

/* If a machine player resigns, it tries to help its friends. */

give_up(side)
Side *side;
{
    Side *side1;

    /* Try to give away all of our units to an ally. */
    for_all_sides(side1) {
	if (side != side1 && allied_side(side, side1)) {
	    resign_game(side, side1);
	    return;
	}
    }
    /* No allies, let everything become neutral. */
    resign_game(side, NULL);
}

/* This function is used by a machine to estimate relative strengths.
   For itself and allies, we can just scan the lists of units, but
   enemy strength must be guessed at. */

/* Should be able to estimate the margin of error also, so as to not
   freak out over inaccurate numbers. */

make_estimates(side)
Side *side;
{
    int u, u2, sn1, x, y;
    short view;
    Side *side1, *side2;
    Strategy *strategy = mplayer(side);
    Unit *unit;

    for_all_sides(side1) {
	sn1 = side_number(side1);
	for_all_unit_types(u) {
	    strategy->strengths[sn1][u] = 0;
	}
	if (side1 == side || allied_side(side, side1)) {
	    for_all_side_units(side1, unit) {
		if (alive(unit)) {  /* probably should test other things */
		    ++(strategy->strengths[sn1][unit->type]);
		}
	    }
	}
    }
    /* Look at the current view to get enemy strength. */
    /* This is too easily faked, and doesn't know about hiding units... */
    /* Should also discount old data. */
    for_all_hexes(x, y) {
	view = unit_view(side, x, y);
	if (view != UNSEEN && view != EMPTY) {
	    side2 = side_n(vside(view));
	    if (enemy_side(side, side2)) {
		++(strategy->strengths[vside(view)][vtype(view)]);
	    }
	}
    }
    /* Estimate how many units of each type in allied group. */
    for_all_sides(side1) {
	sn1 = side_number(side1);
	for_all_unit_types(u) {
	    strategy->alstrengths[sn1][u] = strategy->strengths[sn1][u];
	    for_all_sides(side2) {
		if (side1 != side2 && allied_side(side1, side2)) {
		    strategy->alstrengths[sn1][u] +=
		      strategy->strengths[side_number(side2)][u];
		}
	    }
	}
    }
    /* Dump out a detailed listing of our estimates. */
    if (DebugM) {
	for_all_sides(side1) {
	    DMprintf("%s strength estimate: ", side_desig(side1));
	    for_all_unit_types(u) {
		DMprintf(" %d", strategy->strengths[side_number(side1)][u]);
	    }
	    DMprintf("\n");
	}
    }
}

/* Push a new goalstate onto the side's list of goalstates. */

/* (this should only add goals that are not already present) */

add_goalstate(side, goalstate)
Side *side;
GoalState *goalstate;
{
    goalstate->next = mplayer(side)->goalstates;
    mplayer(side)->goalstates = goalstate;
}

GoalState *
has_goal(side, goaltype)
Side *side;
GoalType goaltype;
{
    GoalState *goalstate;

    for_all_side_goalstates(side, goalstate) {
	if (goalstate->goal->type == goaltype) {
	    return goalstate;
	}
    }
    return NULL;
}

/* This is for when a unit needs a plan and asks its side for one. */

mplayer_decide_plan(side, unit)
Side *side;
Unit *unit;
{
    if (unit->plan) {
	if (mplayer(side)->trytowin) {
	    switch (unit->plan->type) {
	      case UNFORMED_PLAN:
		if (mobile(unit->type)) {
		    /* Maybe assign to exploration. */
		    if (mplayer(side)->searchzones != NULL) {
			if (need_this_type_to_explore(side, unit->type)) {
			    /* also limit to a total percentage, in case
			       exploration needs are very high */
			    assign_to_exploration(side, unit);
			} else {
			    assign_to_offense(side, unit);
			}
		    } else {
			assign_to_offense(side, unit);
		    }
		} else {
		    if (mplayer(side)->searchzones != NULL) {
			if (can_build_needed_explorers(side, unit->type)) {
			    assign_to_explorer_construction(side, unit);
			} else {
		    	    unit->plan->type = DEFENSIVE_PLAN;
		   	    unit->plan->tasks = NULL;
			}
		    } else if (can_build_attackers(side, unit->type)) {
			unit->plan->type = OFFENSIVE_PLAN;
			unit->plan->tasks = NULL;
		    } else {
		    	unit->plan->type = DEFENSIVE_PLAN;
		   	unit->plan->tasks = NULL;
		    }
		}
		break;
	      case FOLLOW_ORDERS:
		/* leave plan alone */
		break;
	      case OFFENSIVE_PLAN:
		/* leave plan alone */
		break;
	      case EXPLORATORY_PLAN:
		/* leave plan alone */
		break;
	      case DEFENSIVE_PLAN:
		/* leave plan alone */
		break;
	    }
	} else {
	    unit->plan->type = RANDOM_PLAN;
	    unit->plan->tasks = NULL;
	}
    }
}

need_this_type_to_explore(side, u)
Side *side;
int u;
{
    return (mobile(u) && flip_coin());
/*    return (mplayer(side)->explorersneeded[u] > mplayer(side)->explorertypes[u]);  */
}

/* Set the unit up as an explorer and let it go. */

assign_to_exploration(side, unit)
Side *side;
Unit *unit;
{
    /* Unit's goal in life will be to see that the world is all known. */
    unit->plan->type = EXPLORATORY_PLAN;
    unit->plan->maingoal = create_goal(WORLD_KNOWN); /* should reuse */
    /* Count it as one of this side's explorers. */
    ++(mplayer(side)->explorertypes[unit->type]);
    DMprintf("%s assigned to exploration\n", unit_desig(unit));
}

can_build_needed_explorers(side, u)
Side *side;
int u;
{
	int u2;
	
	for_all_unit_types(u2) {
		if (mobile(u2) && uu_acp_to_create(u, u2) > 0) return TRUE;
	}
	return FALSE;
}

assign_to_explorer_construction(side, unit)
Side *side;
Unit *unit;
{
	int u = unit->type, u2;
	
    /* Unit's goal in life will be to see that the world is all known. */
    unit->plan->type = EXPLORATORY_PLAN;
/*
	for_all_unit_types(u2) {
		if (mobile(u2) && uu_acp_to_create(u, u2) > 0) {
		}
	}  */
    /* (should) Count it as one of this side's explorer builders. */
    DMprintf("%s assigned to explorer construction\n", unit_desig(unit));
}

assign_to_offense(side, unit)
Side *side;
Unit *unit;
{
	int u = unit->type;

	unit->plan->type = OFFENSIVE_PLAN;
	unit->plan->tasks = NULL;
}

can_build_attackers(side, u)
Side *side;
int u;
{
	int u2;
	
	for_all_unit_types(u2) {
		if (mobile(u2) && uu_acp_to_create(u, u2) > 0) return TRUE;
	}
	return FALSE;
}

/* Returns the type of attack to plan for. */

/*  (Should balance relative effectiveness of each type of attack.) */

attack_type(e)
int e;
{
#if 0
    int u;

    if (u_surrender(e) > 0 || u_siege(e) > 0) return BESIEGE;
    for_all_unit_types(u) if (could_capture(u, e)) return CAPTARGET;
    return HITTARGET;
#endif
}

