/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact.  */

/* Event handling for the X11 version of xconq. */

#include "conq.h"
#include "xconq.h"

#include <sys/types.h>
#include <sys/time.h>

/* All input comes in through this function. */

/* (This code should use fd masks...) */

get_input(tmout)
int tmout;
{
    int pending = FALSE, gotinput = FALSE;
    int i, width = 32 /* getdtablesize() */;
    fd_set readfds;
    struct timeval timeout;
    Side *side, *asides[32]; /* array size should match "width" */

    /* Build up the mask of X connection fd's to multiplex on.  Also see
       if anybody already has pending events; if so, process them and return
       directly, don't bother with select().  We should get back here
       very shortly (implies no long loops elsewhere in code). */
    FD_ZERO(&readfds);
    for_all_sides(side) {
	if (active_display(side)) {
	    FD_SET(ConnectionNumber(side->ui->dpy), &readfds);
	    asides[ConnectionNumber(side->ui->dpy)] = side;
	    if (XPending(side->ui->dpy) > 0) {
		process_events(side);
		pending = TRUE;
	    } 
	}
    }
    if (pending) {
	DGprintf("Found and processed already-pending events\n");
	return TRUE;
    }
    /* OK, now we know all the pending events are gone; we now want to
       hang around waiting for more, possibly forever. */
    /* First make sure each player knows that input is possible, and
       maybe switch to an appropriate unit. */
    for_all_sides(side) {
	if (active_display(side)) {
	    Map *map;

	    for_all_maps(side, map) {
		if (map->curtool == UNITMOVETOOL
		    && map->curunit != NULL
		    && map->curunit->act != NULL
		    && (map->curunit->act->acp == 0
			|| map->curunit->act->mp == 0)) {
		    set_map_curunit(side, map,
				    find_next_actor(side, map->curunit));
		}
	    }
	}
    }
    if (DebugG) {
	DGprintf("Waiting %d msecs for input from ", tmout);
	for_all_sides(side) {
	    if (active_display(side)) DGprintf("%s ", side_desig(side));
	}
	DGprintf("\n");
    }
    if (tmout >= 0) {
	timeout.tv_sec = tmout / 1000;
	timeout.tv_usec = tmout % 1000;
    } else {
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
    }
    if (select(width, &readfds, NULL, NULL,
	       (tmout >= 0 ? &timeout : NULL)) < 0) {
	fprintf(stderr, "error in select!\n");
	abort();
    } else {
	/* Pick out the sides that actually sent events. */
	/* (This does one event per side, then returns...) */
	for (i = 0; i < width; ++i) {
	    /* Snapshot the time first, in case event processing is slow. */
	    if (FD_ISSET(i, &readfds)) {
		gotinput = TRUE;
		asides[i]->lasttime = time(0);
	    }
	    if (FD_ISSET(i, &readfds)) {
		process_events(asides[i]);
	    }
	}
    }
    return gotinput;
}

/* Look at a single event and react appropriately. */

/* (This needs to check on timeouts?) */

/* (need to be able to change the "front" map) */

process_events(side)
Side *side;
{
    char buf[BUFSIZE], ch;
    int nchar, rawx, rawy;
    Map *map;
    XEvent evt;

    XNextEvent(side->ui->dpy, &evt);
    DGprintf("%s gets an event, type %d\n", side_desig(side), evt.type);
    switch (evt.type) {
      case KeyPress:
	nchar = XLookupString(&(evt.xkey), buf, BUFSIZE, NULL, NULL);
	if (nchar > 0) {
	    ch = buf[0];
	    DGprintf("%s returns char '%c'\n", side_desig(side), ch);
	    /* so we can do ^L even while typing a string */
	    if (ch == '\014') { /* control l */
		redraw(side);
#if 0
	    } else if (evt.xbutton.window == side->ui->boolpopup) {
		if (handle_bool_input(side, ch)) {
		    if (side->ui->modalhandler) {
			int (*fn)() = side->ui->modalhandler;

			/* Take out handler, will restore itself if needed. */
			side->ui->modalhandler = NULL;
			(*fn)(side, (ch == '\033'));
		    } else {
			/* nothing to do with the bool? */
		    }
		}
	    } else if (evt.xbutton.window == side->ui->stringpopup) {
		if (handle_string_input(side, ch)) {
		    if (side->ui->modalhandler) {
			int (*fn)() = side->ui->modalhandler;

			/* Take out the handler, will restore itself if needed. */
			side->ui->modalhandler = NULL;
			(*fn)(side, (ch == '\033'));
		    } else {
			/* nothing to do with the string? */
		    }
		}
#endif
	    } else {
		side->ui->reqch = ch;
		/* Call the handler, giving it side and cancel flag. */
		if (side->ui->modalhandler) {
		    int (*fn)() = side->ui->modalhandler;

		    /* Take out the handler, will restore itself if needed. */
		    side->ui->modalhandler = NULL;
		    (*fn)(side, (ch == '\033'));
		} else {
		    interp_general_command(side);
		}
	    }
	} else {
	    /* pretty strange, ok to do nothing? */
	}
	break;
      case ButtonPress:
	if (button_down_map(side, evt)) {
	} else {
	    DGprintf("Button press on unhandled window\n");
	}
	break;
      case ButtonRelease:
	if (button_up_map(side, evt)) {
	} else {
	    DGprintf("Button release on unhandled window\n");
	}
	break;
      case Expose:
	/* Handle the expose events.  This code needs to be careful not to
	   do too many redraws, since some of them can be quite slow. */
	if (evt.xexpose.count == 0) {
	    if (evt.xexpose.window == side->ui->help) {
		/* show help */
	    } else if (expose_map(side, evt)) {
	    } else {
		DGprintf("Exposure of unhandled window\n");
	    }
	}
	/* should flush other events here? */
	break;
      case DestroyNotify:
	run_warning("want to destroy a window");
	break;
      case ConfigureNotify:
	/* Need to recalculate viewports if map windows change size. */
/*	do { */
	    for_all_maps(side, map) {
		if (evt.xconfigurerequest.window == map->window) {
		    resize_map(side, map,
			       evt.xconfigurerequest.width,
			       evt.xconfigurerequest.height);
		    break;
		}
	    }
/*	} while (XCheckMaskEvent(side->ui->dpy, StructureNotifyMask, &evt));*/
	/* Other win types don't need anything special, just fall through. */
	break;
      case MapNotify:
	/* Don't need to do anything here. */
	break;
      default:
	/* X might send us events we don't care about, so don't get bent
	   out of shape. */
        DGprintf("Unknown X event type %d, ignoring", evt.type);
	break;
    }
    return FALSE;
}

#if 0
    /* some commands OK, so where should this go? */
    if (period_giventime() && !side->timedout && side->timeleft <= 0) {
	side->timedout = TRUE;
	notify(side, "You ran out of time!!");
	beep(side);
    }
#endif

int tmpx, tmpy;

/* This interprets generic commands. */

interp_general_command(side)
Side *side;
{
    int dir, terr;

    DGprintf("%s keyboard input: %c (%d)\n",
	     side->name, side->ui->reqch, side->ui->reqvalue);
    if (isdigit(side->ui->reqch)) {
	if (side->ui->reqvalue == DFLTFLAG) {
	    side->ui->reqvalue = 0;
	} else {
	    side->ui->reqvalue *= 10;
	}
	side->ui->reqvalue +=
	  (side->ui->reqvalue >= 0 ? 1 : -1) * (side->ui->reqch - '0');
	return;
    } else {
	execute_command(side, side->ui->reqch, side->ui->reqvalue);
	side->ui->reqvalue = DFLTFLAG;
    }
}

/* Prompt for a type of a unit from player, maybe only allowing some types
   to be accepted.  Also allow specification of no unit type.  We do this
   by scanning the vector, building a string of chars and a vector of
   unit types, so as to be able to map back when done. */

ask_unit_type(side, prompt, possibles)
Side *side;
char *prompt;
int *possibles;
{
    char tmpbuf[100];
    char ucharbuf[MAXUTYPES+1];
    int i, numtypes = 0, u, type;

    for_all_unit_types(u) {
	if (possibles == NULL || possibles[u]) {
	    ucharbuf[numtypes++] = utype_name_n(u, 1)[0];
	    numtypes++;
	}
    }
    ucharbuf[numtypes] = '\0';
    if (numtypes == 0) {
	return TRUE;
    } else {
	sprintf(tmpbuf, "%s [%s]", prompt, ucharbuf);
	ask_string(side, tmpbuf, "");
	return TRUE;
    }
}

/* Do something with the char or unit type that the player entered. */

grok_unit_type(side)
Side *side;
{
    int i, type = -1;

    type = side->ui->reqch;

    return type;
}

ask_terrain_type(side, prompt, possibles)
Side *side;
char *prompt;
short *possibles;
{
    int numtypes = 0, t;
    char tmpbuf[BUFSIZE];
    char tcharbuf[MAXTTYPES+1];

    for_all_terrain_types(t) {
	side->ui->bvec[t] = 0;
	if (possibles == NULL || possibles[t]) {
	    side->ui->bvec[t] = 1;
	    side->ui->uvec[numtypes] = t;
/*	    tcharbuf[numtypes] = t_char(t); */
	    numtypes++;
	}
    }
    tcharbuf[numtypes] = '\0';
    if (numtypes == 0) {
	return FALSE;
    } else {
	sprintf(tmpbuf, "%s [%s]", prompt, tcharbuf);
	ask_string(side, tmpbuf, "");
	return TRUE;
    }
}

/* Do something with the char or terrain type that the player entered. */

grok_terrain_type(side)
Side *side;
{
    return 0;
}

/* User is asked to pick a position on map.  This will iterate until the */
/* space bar designates the final position. */

/* (should change the cursor temporarily) */

ask_position(side, prompt)
Side *side;
char *prompt;
{
    sprintf(tmpbuf, "%s [click to set]", prompt);
    ask_string(side, tmpbuf, "");
    /* need to change mode or somehting */
}

/* Restore the saved "cur" slots. */

restore_cur(side)
Side *side;
{
}

grok_position(side)
Side *side;
{
    return FALSE;
}

/* Prompt for a yes/no answer with a settable default. */

ask_bool(side, question, dflt)
Side *side;
char *question;
int dflt;
{
    sprintf(side->ui->prompt, "%s", question);
}

handle_bool_input(side, ch)
Side *side;
char ch;
{
    int len = strlen(side->ui->answer);

    if (ch == 'y' || ch == 'n') {
	side->ui->reqvalue2 = ch;
	return TRUE;
    }
    return FALSE;
}

/* Figure out what the answer actually is, keeping the default in mind. */

grok_bool(side)
Side *side;
{
#if 0
    int dflt = side->ui->reqvalue2;
    char ch = side->ui->stringanswer[0];

    if (dflt ? (lowercase(ch) == 'n') : (lowercase(ch) == 'y'))
      dflt = !dflt;
    return dflt;
#endif
    return (side->ui->reqvalue2 == 'y');
}

/* Read a string from the prompt window.  Deletion is allowed, and a
   text cursor (an underscore) is displayed. */

ask_string(side, prompt, dflt)
Side *side;
char *prompt, *dflt;
{
    sprintf(side->ui->prompt, "%s", prompt);
    if (dflt == NULL) dflt = "";
    sprintf(side->ui->answer, "%s", dflt);
}

handle_string_input(side, ch)
Side *side;
char ch;
{
    int len = strlen(side->ui->answer);

    if (ch == '\r' || ch == '\n') {
	return TRUE;
    } else {
	if (ch == BACKSPACE) {
	    if (len > 0) --len;
	} else {
	    side->ui->answer[len++] = ch;
	}
	side->ui->answer[len] = '\0';
	show_string_popup(side);
    }
    return FALSE;
}

/* Dig a character from the filled-in request and add it into the string. */
/* Keep returning NULL until we get something. */

char *
grok_string(side)
Side *side;
{
    return side->ui->answer;
}

/* Open or reuse a unit closeup. */

make_current(side, unit)
Side *side;
Unit *unit;
{
#if 0
    Map *map;

    if (active_display(side) && unit != NULL) {
	if (alive(unit)) {
	    if (side->ui->closeups == NULL) {
		create_closeup(side, unit, TRUE);
	    }
	    if (side->ui->closeups == NULL) {
		fprintf(stderr, "No closeup made!\n");
		return;
	    } else {
		/* not quite right */
		side->ui->closeups->unit = unit;
		show_closeup(side, side->ui->closeups);
	    }
	    /* a little too drastic probably */
	    for_all_maps(side, map) {
		map->curunit = unit;
		show_current_info(side, map);
	    }
	} else {
	    beep(side);
	}
    }
#endif
}

/* When doing a survey, you can move the cursor anywhere and it will show */
/* what is at that hex, but not give away too much! */

move_look(side, map, nx, ny)
Side *side;
Map *map;
int nx, ny;
{
    Unit *unit;

    if (inside_area(nx, ny)) {
	erase_cursor(side, map);
	if (!in_view(side, nx, ny)) beep(side);
	unit = unit_at(nx, ny);
	if (unit != NULL && (unit->side == side || side->designer)) {
	    set_map_curunit(side, map, unit);
	} else {
	    side->ui->curunit = map->curunit = NULL;
	    show_current_info(side, map);
	}
    } else {
	if (active_display(side)) beep(side);
    }
}

/* Set the "current" unit of a side - the one being displayed, moved, etc. */

set_map_curunit(side, map, unit)
Side *side;
Map *map;
Unit *unit;
{
    if (active_display(side)) {
	if (unit != NULL
	    && alive(unit)
	    && (unit->side == side
		|| allied_side(unit->side,side)
		|| Debug
		|| numdesigners > 0)) {
	    side->ui->curunit = map->curunit = unit;
	    show_current_info(side, map);
	} else {
	    side->ui->curunit = map->curunit = NULL;
	    erase_cursor(side, map);
	    show_current_info(side, map);  /* maybe not? */
	}
    }
}

show_current_info(side, map)
Side *side;
Map *map;
{
    draw_cursor(side, map);
    show_map_info(side, map);
}

/* See if the button event happened in a map display, record where etc,
   then return true. */

button_down_map(side, evt)
Side *side;
XEvent evt;
{
    Map *map;
    int rawx, rawy;

    for_all_maps(side, map) {
	if (evt.xbutton.window == map->viewwin) {
	    rawx = evt.xbutton.x;  rawy = evt.xbutton.y; 
	    /* Remember which map this happened on. */
	    side->ui->downmap = map;
	    /* Get area coords and closest hex side. */
	    deform(side, map, rawx, rawy,
		   &(side->ui->downx), &(side->ui->downy),
		   &(side->ui->downdir));
	    DGprintf("%s returns map down %d %d dir %d\n",
		     side_desig(side),
		     side->ui->downx, side->ui->downy, side->ui->downdir);
	    return TRUE;
	} else if (evt.xbutton.window == map->infowin) {
	    return TRUE;
	} else if (evt.xbutton.window == map->controls) {
	    /* (should collect and use shift etc) */
	    handle_control_click(side, map,
				 evt.xbutton.x / (conwid / 2),
				 evt.xbutton.y / conhgt);
	    return TRUE;
	} else if (evt.xbutton.window == map->vscrollbar) {
	    scroll_map_absolute(side, map, -1, evt.xbutton.y);
	    return TRUE;
	} else if (evt.xbutton.window == map->hscrollbar) {
	    scroll_map_absolute(side, map, evt.xbutton.x, -1);
	    return TRUE;
	} else if (evt.xbutton.window == map->window) {
	    return TRUE;
	}
    }
    return FALSE;
}

/* See if the button event happened in a map display, record where etc,
   then return true. */

button_up_map(side, evt)
Side *side;
XEvent evt;
{
    Map *map;
    int rawx, rawy, xx;

    for_all_maps(side, map) {
	if (evt.xbutton.window == map->viewwin) {
	    side->ui->upmap = map;
	    if (side->ui->upmap == side->ui->downmap) {
		rawx = evt.xbutton.x;  rawy = evt.xbutton.y; 
		/* Get area coords and closest hex side. */
		deform(side, map, rawx, rawy,
		       &(side->ui->upx), &(side->ui->upy),
		       &(side->ui->updir));
		DGprintf("%s returns map 0x%x up %d %d dir %d\n",
			 side_desig(side), map,
			 side->ui->upx, side->ui->upy, side->ui->updir);
		if (side->ui->downx == side->ui->upx
		    && side->ui->downy == side->ui->upy) {
		    handle_map_click(side, map);
		} else {
		    handle_map_drag(side, map);
		}
	    } else {
		/* dragged to random place, forget about it */
	    }
	    return TRUE;
	} else if (evt.xbutton.window == map->infowin) {
	    return TRUE;
	} else if (evt.xbutton.window == map->controls) {
	    /* Action was done on button down. */
	    return TRUE;
	} else if (evt.xbutton.window == map->vscrollbar) {
	    /* Action was done on button down. */
	    return TRUE;
	} else if (evt.xbutton.window == map->hscrollbar) {
	    /* Action was done on button down. */
	    return TRUE;
	} else if (evt.xbutton.window == map->window) {
	    return TRUE;
	}
    }
    return FALSE;
}

int clicked_tool();
int clicked_ttype();
int clicked_utype();
int clicked_curside();
int clicked_zoom();
int clicked_style();

/* The list of all the types of control panel items. */

Controltype controltypes[NUMCONTROLS] = {
    { LOOK, "look", clicked_tool, LOOKTOOL, FALSE },
    { 0, NULL, NULL, -1, FALSE },

    { UNITMOVE, "move", clicked_tool, UNITMOVETOOL, FALSE },
    { 0, NULL, NULL, -1, FALSE },

    { UNITSHOOT, "shoot", clicked_tool, UNITSHOOTTOOL, FALSE },
    { 0, NULL, NULL, -1, FALSE },

    { UNITBUILD, "build", clicked_tool, -1 /* UNITBUILDTOOL */, FALSE },
    { 0, NULL, NULL, -1, FALSE },

    { 0, NULL, NULL, -1, FALSE },
    { 0, NULL, NULL, -1, FALSE },

    { SHOWTERR, "showterr", clicked_style, -1, FALSE },
    { 0, NULL, NULL, -1, FALSE },
    { SHOWUNITS, "showunits", clicked_style, -1, FALSE },
    { 0, NULL, NULL, -1, FALSE },
    { SHOWPATS, "showpats", clicked_style, -1, FALSE },
    { 0, NULL, NULL, -1, FALSE },
    { SHOWNAMES, "shownames", clicked_style, -1, FALSE },
    { 0, NULL, NULL, -1, FALSE },
    { SHOWEDGES, "showedges", clicked_style, -1, FALSE },
    { 0, NULL, NULL, -1, FALSE },
    { SHOWPEOPLE, "showpeople", clicked_style, -1, FALSE },
    { 0, NULL, NULL, -1, FALSE },
    { SHOWELEVS, "showelevs", clicked_style, -1, FALSE },
    { 0, NULL, NULL, -1, FALSE },
    { SHOWTEMP, "showtemp", clicked_style, -1, FALSE },
    { 0, NULL, NULL, -1, FALSE },

    { ZOOMOUT, "zoom out", clicked_zoom, -1, FALSE },
    { ZOOMIN, "zoom in", clicked_zoom, -1, FALSE },

    { 0, NULL, NULL, -1, FALSE },
    { 0, NULL, NULL, -1, FALSE },

#ifdef DESIGNERS
    { HEXPAINT, "hex", clicked_tool, HEXPAINTTOOL, TRUE },
    { CURTTYPE, "curttype", clicked_ttype, -1, TRUE },

    { UNITADD, "unit", clicked_tool, UNITADDTOOL, TRUE },
    { CURUTYPE, "curutype", clicked_utype, -1, TRUE },

    { CURSIDE, "curside", clicked_curside, -1, TRUE },
    { 0, NULL, NULL, -1, FALSE }
#endif /* DESIGNERS */
};

/* Clicking on a tool selects and changes the current tool. */

clicked_tool(side, map, which)
Side *side;
Map *map;
int which;
{
    int i;

    for (i = 0; i < NUMCONTROLS; ++i) {
	if (controltypes[i].tool == map->curtool) {
	    map->controlbs[i].selected = FALSE;
	    break;
	}
    }
    map->curtool = controltypes[which].tool;
    map->controlbs[which].selected = TRUE;
    set_tool_cursor(side, map);
    return TRUE;
}

/* If a given tool (mode) has a special cursor, use it, otherwise go back
   to the look tool's cursor. */

set_tool_cursor(side, map)
Side *side;
Map *map;
{
    int tool;

    if (side->ui->toolcursors[map->curtool] != None) {
	tool = map->curtool;
    } else {
	/* revert to default cursor */
	tool = LOOKTOOL;
    }
    XDefineCursor(side->ui->dpy, map->viewwin, side->ui->toolcursors[tool]);
}

#ifdef DESIGNERS

/* If clicked on the terrain type display, go to the next type. */
/* (should have something to go back too) */

clicked_ttype(side, map, which)
Side *side;
Map *map;
int which;
{
    map->curttype = (map->curttype + 1) % numttypes;
    map->controlbs[which].color = side->ui->hexcolor[map->curttype];
    /* Should do pic also, esp for noxxx cases */
    /* could do meaningful pics for linears also */
    return TRUE;
}

/* Cycle around all the unit types. */

clicked_utype(side, map, which)
Side *side;
Map *map;
int which;
{
    map->curutype = (map->curutype + 1) % numutypes;
    return TRUE;
}

/* Cycle around possible sides, including indeps as side 0. */

clicked_curside(side, map, which)
Side *side;
Map *map;
int which;
{
    map->cursidenumber = (map->cursidenumber + 1) % numsides;
    /* do emblems instead? */
    sprintf(map->controlbs[which].text, "%s",
	    side_desig(side_n(map->cursidenumber)));
    return TRUE;
}
#endif /* DESIGNERS */

/* Clicked on a zoom button. */

clicked_zoom(side, map, which)
Side *side;
Map *map;
int which;
{
    if (which == ZOOMOUT) {
	zoom_out(side, map);
    } else {
	zoom_in(side, map);
    }
    map->controlbs[ZOOMOUT].enabled = (map->power > 0);
    map->controlbs[ZOOMIN].enabled = (map->power < NUMPOWERS-1);
    draw_controls(side, map);
    return FALSE;
}

/* these should disable off-the-end zooms. */

zoom_in(side, map)
Side *side;
Map *map;
{
    if (map->power < NUMPOWERS - 1) {
	++(map->power);
	set_map_power(map, map->power);
	set_map_viewport(map);
	show_map(side, map);
    }
}

zoom_out(side, map)
Side *side;
Map *map;
{
    if (map->power > 0) {
	--(map->power);
	set_map_power(map, map->power);
	set_map_viewport(map);
	show_map(side, map);
    }
}

/* Clicking on a display style usually toggles it. */

clicked_style(side, map, which)
Side *side;
Map *map;
int which;
{
    int state;

    switch (which) {
      case SHOWTERR:
	state = map->drawterrain = !map->drawterrain;
	break;
      case SHOWPATS:
	state = map->drawhexpats = !map->drawhexpats;
	break;
      case SHOWEDGES:
	state = map->drawgrid = !map->drawgrid;
	break;
      case SHOWUNITS:
	state = map->drawunits = !map->drawunits;
	break;
      case SHOWNAMES:
	state = map->drawnames = !map->drawnames;
	break;
      case SHOWPEOPLE:
	state = map->drawpeople = !map->drawpeople;
	break;
      case SHOWELEVS:
	state = map->drawelevations = !map->drawelevations;
	break;
      case SHOWTEMP:
	state = map->drawtemp = !map->drawtemp;
	break;
      default:
	break;
    }
    /* Tweak the control to reflect the correct internal state. */
    map->controlbs[which].selected = state;
    /* change the text of text buttons perhaps */
    /* also turn off others? */
    /* display the effect */
    show_map(side, map);
    return FALSE;
}

/* Invoke the appropriate action for a control panel item. */

handle_control_click(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
    int which = 2 * y + x;

    if (x == 1 && controltypes[which].name == NULL) --which;

    if (map->controlbs[which].enabled) {
	if (controltypes[which].fn != NULL) {
	    if ((*(controltypes[which].fn))(side, map, which)) {
		draw_controls(side, map);
	    }
	} else {
	    run_warning("Undefined control item!");
	}
    } else {
	/* ignore, don't even beep */
    }
}

/* Put the map at an absolute position. */

scroll_map_absolute(side, map, sx, sy)
Side *side;
Map *map;
int sx, sy;
{
    if (sx >= 0) {
	/* Only do if viewport is smaller than area. */
	if (area.xwrap || map->vw < area.width) {
	    map->vcx = ((area.width * sx) / map->pxw);
	}
    }
    if (sy >= 0) {
	if (map->vh < area.height) {
	    map->vcy = (((area.height * (map->pxh - sy)) / map->pxh));
	}
    }
    /* Now adjust the viewport position as needed. */
    set_map_viewport(map);
    show_map(side, map);
}

/* A click is when the pointer went down and up in the same place. */

handle_map_click(side, map)
Side *side;
Map *map;
{
    if (map == NULL) { beep(side); return; }
    switch (map->curtool) {
      case LOOKTOOL:
	move_look(side, map, side->ui->downx, side->ui->downy);
	break;
      case UNITMOVETOOL:
	if (map->curunit) {
	    /* Tell the selected unit to go to the clicked-on cell. */
	    order_moveto(map->curunit, side->ui->downx, side->ui->downy);
	} else {
	    /* is this sensible? */
	    move_look(side, map, side->ui->downx, side->ui->downy);
	}
	break;
      case UNITSHOOTTOOL:
	if (map->curunit) {
	    Unit *unit2;

	    if (inside_area(side->ui->downx, side->ui->downy)) {
		if ((unit2 = unit_at(side->ui->downx, side->ui->downy)) != NULL) {
		    if (map->curunit != unit2) {
			do_attack_action(map->curunit, unit2, 100);
		    } else {
			/* don't attack ourselves */
		    }
		} else {
		    /* We're just shooting for the hell of it. */
		    do_fire_into_action(map->curunit, side->ui->downx, side->ui->downy, 0, 0);
		}
	    } else {
		beep(side);
	    }
	} else {
	}
	break;
#ifdef DESIGNERS
      case HEXPAINTTOOL:
	switch (t_subtype(map->curttype)) {
	  case cellsubtype:
	    paint_terrain(side,
			  side->ui->downx, side->ui->downy, 0, 0,
			  map->curttype);
	    break;
	  case bordersubtype:
	    paint_border(side,
			 side->ui->downx, side->ui->downy, side->ui->downdir,
			 map->curttype);
	    break;
	  case connectionsubtype:
	    paint_connection(side,
			   side->ui->downx, side->ui->downy, side->ui->downdir,
			     map->curttype);
	    break;
	  case coatingsubtype:
	    break;
	  default:
	    case_panic("terrain subtype", t_subtype(map->curttype));
	    break;
	}
	break;
      case UNITADDTOOL:
	add_unit(side, side->ui->downx, side->ui->downy,
		 map->curutype, side_n(map->cursidenumber));
	break;
#endif /* DESIGNERS */
      default:
	/* error eventually */
	break;
    }
}

handle_map_drag(side, map)
Side *side;
Map *map;
{	    
    if (map == NULL) { beep(side); return; }
    switch (map->curtool) {
      case LOOKTOOL:
	if (map->curunit) {
	    /* (a trick to avoid switching modes) */
	    order_moveto(map->curunit, side->ui->upx, side->ui->upy);
	} else {
	    move_look(side, map, side->ui->downx, side->ui->downy);
	}
	break;
      case UNITMOVETOOL:
	if (map->curunit) {
	    /* or order a patrol? */
	    order_moveto(map->curunit, side->ui->upx, side->ui->upy);
	}
	break;
      case UNITSHOOTTOOL:
	/* what would this mean? */
	break;
#ifdef DESIGNERS
      case HEXPAINTTOOL:
	switch (t_subtype(map->curttype)) {
	  case cellsubtype:
	    paint_terrain(side, side->ui->downx, side->ui->downy,
			  abs(side->ui->upx - side->ui->downx),
			  abs(side->ui->upy - side->ui->downy),
			  map->curttype);
	    break;
	  case bordersubtype:
/*	    paint_border_chain(side,
			      side->ui->downx, side->ui->downy,
			      side->ui->upx, side->ui->upy,
			      map->curttype); */
	    break;
	  case connectionsubtype:
/*	    paint_connection_chain(side,
				  side->ui->downx, side->ui->downy,
				  side->ui->upx, side->ui->upy,
				  map->curttype); */
	    break;
	  case coatingsubtype:
	    break;
	  default:
	    case_panic("terrain subtype", t_subtype(map->curttype));
	    break;
	}
	break;
      case UNITADDTOOL:
	add_unit(side, side->ui->downx, side->ui->downy,
		 map->curutype, side_n(map->cursidenumber));
	break;
#endif /* DESIGNERS */
      default:
	/* error eventually */
	break;
    }
}

/* See if the expose event is doing some or all of a map. */

expose_map(side, evt)
Side *side;
XEvent evt;
{
    Map *map;

    for_all_maps(side, map) {
	if (evt.xexpose.window == map->viewwin) {
	    /* should be more efficient and redraw only exposed part? */
	    show_map_view(side, map);
	    return TRUE;
	} else if (evt.xexpose.window == map->msg) {
	    show_note(side, map);
	    return TRUE;
	} else if (evt.xexpose.window == map->game) {
	    show_game_state(side, map);
	    return TRUE;
	} else if (evt.xexpose.window == map->sides) {
	    show_sides(side, map);
	    return TRUE;
	} else if (evt.xexpose.window == map->list) {
	    return TRUE;
	} else if (evt.xexpose.window == map->infowin) {
	    show_map_info(side, map);
	    return TRUE;
	} else if (evt.xexpose.window == map->controls) {
	    show_controls(side, map);
	    return TRUE;
	} else if (evt.xexpose.window == map->vscrollbar) {
	    /* should only do the one... */
	    draw_scrollbars(side, map);
	    return TRUE;
	} else if (evt.xexpose.window == map->hscrollbar) {
	    draw_scrollbars(side, map);
	    return TRUE;
	} else if (evt.xexpose.window == map->window) {
	    /* Don't need to do anything here, because the subwindows
	       are supposed to tile, and even if they don't, the
	       exposes of subwindows will do any necessary redrawing. */
	    return TRUE;
	}
    }
    return FALSE;
}

#if 0

/* Interpret input to the doctrine dialog. */

command_docdial(side, ch)
Side *side;
char ch;
{
    switch (ch) {
      case '+':
	++(side->doctrine->realtimeout);
	break;
      case '-':
	--(side->doctrine->realtimeout);
	break;
#if 0
      case 'r':
	side->humanp = !(side->humanp);
	if (side->humanp) {
	    numhumans++;
	    side->strategy = NULL;  /* should free all bits? */
	    notify(side, "Strategy returned to human control.");
	} else {
	    numhumans--;
	    /* Give the next unit something to work from. */
	    init_machine_turn(side);
	    notify(side, "Robot code now doing strategy.");
	}
#endif
	break;
      default:
	beep(side);
    }
    show_our_closeup(side);
}

button_down_docdial(side, rawx, rawy)
Side *side;
int rawx, rawy;
{
    side->doctrine->unitseveraskside = !side->doctrine->unitseveraskside;
    show_our_closeup(side);
}
#endif

/* Get rid of extra input events, generally because the situation has
   changed drastically and they are no longer of interest.  This routine
   has relatively few valid uses. */

flush_input(side)
Side *side;
{
    printf("doing an input flush\n");
    if (active_display(side)) {
	XSync(side->ui->dpy, TRUE);  
    }
}

/* Trivial abstraction - sometimes other routines need to ensure all output
   so far is actually on the screen and not being buffered up somewhere. */

flush_output(side) 
Side *side; 
{  
    if (active_display(side)) XFlush(side->ui->dpy);  
}

#ifdef DEBUGGING
/* Set all the window backgrounds, borders, GCs, etc, usually to reflect
   inversion of foreground and background. */

reset_color_state(side)
Side *side;
{
    XGCValues values;
    unsigned int gcmask = GCForeground | GCBackground;

#if 0
    /* Set all the window backgrounds. */
    XSetWindowBackground(side->ui->dpy, side->ui->msg, side->ui->bgcolor);
    XSetWindowBorder(side->ui->dpy, side->ui->msg, side->ui->fgcolor);
#endif

    values.foreground = side->ui->fgcolor;
    values.background = side->ui->bgcolor;
    XChangeGC(side->ui->dpy, side->ui->gc, gcmask, &values);
    XChangeGC(side->ui->dpy, side->ui->textgc, gcmask, &values);
    XChangeGC(side->ui->dpy, side->ui->ltextgc, gcmask, &values);
    XChangeGC(side->ui->dpy, side->ui->icongc, gcmask, &values);
    XChangeGC(side->ui->dpy, side->ui->unitgc, gcmask, &values);
    values.foreground = side->ui->bgcolor;
    values.background = side->ui->fgcolor;
    XChangeGC(side->ui->dpy, side->ui->invicongc, gcmask, &values);
}
#endif /* DEBUGGING */

/* Center the screen on the current location. */

do_recenter(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
    if (map != NULL) {
	recenter(side, map, x, y);
    } else {
	cmd_error(side, "No map!");
    }
}
