/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* This file implements the various textual displays of xconq. */

#include "conq.h"
#include "xconq.h"

char tmpnbuf[BUFSIZE];          /* tmp buf for notices only. */

/* The kernel calls this to update info about the given side. */

update_side_display(side, side2, rightnow)
Side *side, *side2;
int rightnow;
{
    Map *map;

    if (active_display(side)) {
	for_all_maps(side, map) {
	    draw_side_info(side, map, side2);
	}
	if (rightnow) flush_output(side);
    }
}

/* The kernel calls this to update info about the given unit. */

update_unit_display(side, unit, rightnow)
Side *side;
Unit *unit;
int rightnow;
{
    Map *map;

    if (active_display(side)) {
	/* Redraw any/all info about unit. */
	for_all_maps(side, map) {
	    show_map_info(side, map);
	}
	if (rightnow) flush_output(side);
    }
}

update_action_result_display(side, unit, rslt, rightnow)
Side *side;
Unit *unit;
int rslt, rightnow;
{
    if (active_display(side)) {
	notify(side, "%s %s %s!", unit_desig(unit),
	       action_desig(&(unit->act->nextaction)), hevtdefns[rslt].name);
    }
}

char *notices[1];

/* Send a message to everybody who's got a screen. */

/*VARARGS*/
notify_all(control, a1, a2, a3, a4, a5, a6)
char *control, *a1, *a2, *a3, *a4, *a5, *a6;
{
    Side *side;

    for_all_sides(side) notify(side, control, a1, a2, a3, a4, a5, a6);
}

/*VARARGS*/
notify(side, control, a1, a2, a3, a4, a5, a6, a7, a8)
Side *side;
char *control, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8;
{
    int i;
    HistEvent *hevt;

    if (active_display(side)) {
	sprintf(tmpnbuf, control, a1, a2, a3, a4, a5, a6, a7, a8);
	/* Capitalize first char of notice. */
	if (islower(tmpnbuf[0])) tmpnbuf[0] = toupper(tmpnbuf[0]);
	hevt = record_event(-1);
	hevt->data = copy_string(tmpnbuf);
	notices[0] = copy_string(tmpnbuf);
	DGprintf("To %s: %s\n", side_desig(side), tmpnbuf);
    }
}

/* Notice area refresher. */

show_note(side, map)
Side *side;
Map *map;
{
    if (notices[0] != NULL) {
	draw_fg_text(side, map->msg,
		     margin + 1 * side->ui->fw, 1 * side->ui->fh, notices[0]);
    }
}

/* Describe the state of the given unit, in maximal detail. */

show_map_info(side, map)
Side *side;
Map *map;
{
    int i, u, r, nums[MAXUTYPES], xpos;
    Unit *unit = map->curunit, *occ, *mainunit;

    if (map->curunit == NULL) {
	draw_info_text(side, map, 0, 0, 10, "(no unit)");
	return;
    }
    u = unit->type;
    /* Say which unit this is. */
    sprintf(spbuf, "%s", unit_handle(side, unit));
    draw_info_text(side, map, 0, 0, 30, spbuf);
    /* Mention its transport if there is one. */
    sprintf(spbuf, "");
    if (unit->transport != NULL) {
	sprintf(spbuf, "In %s", short_unit_handle(unit->transport));
    }
    draw_info_text(side, map, 0, 1, 40, spbuf);
    /* Very briefly list the numbers and types of the occupants. */
    sprintf(spbuf, "");
    if (unit->occupant != NULL) {
	strcpy(spbuf, "Occ ");
	for_all_unit_types(i) nums[i] = 0;
	for_all_occupants(unit, occ) ++nums[occ->type];
	for_all_unit_types(i) {
	    if (nums[i] > 0) {
		sprintf(tmpbuf, "%d %1s  ", nums[i], utype_name_n(i, 1));
		strcat(spbuf, tmpbuf);
	    }
	}
    }
    draw_info_text(side, map, 0, 2, 40, spbuf);
    /* Describe the "important" parameters like hit points and moves. */
    /* (should say something about parts?) */
    sprintf(spbuf, "HP %d", unit->hp);
    if (unit->hp < u_hp(u)) {
	sprintf(tmpbuf, "/%d", u_hp(u));
	strcat(spbuf, tmpbuf);
    }
    if (unit->cp < u_cp(u) && u_cp(u) > 0) {
	sprintf(tmpbuf, "  %d%% complete", (unit->cp * 100) / u_cp(u));
	strcat(spbuf, tmpbuf);
    } else {
	if (u_acp(u) > 0) {
	    sprintf(tmpbuf, "  Ac %d/%d", unit->act->acp, unit->act->initacp);
	    strcat(spbuf, tmpbuf);
	}
	if (u_speed(u) > 0) {
	    sprintf(tmpbuf, "  Mv %d/%d", unit->act->mp, unit->act->initmp);
	    strcat(spbuf, tmpbuf);
	}
    }
    draw_info_text(side, map, 30, 0, -1, spbuf);
#if 0
    /* Describe the state of construction. */
    if (building(unit)) {
	sprintf(spbuf, "Will complete %s in ???",
		unit_handle(unit->plan->abuilding));
	if (unit->plan->nextbuild == unit->plan->abuilding->type) {
	    sprintf(tmpbuf, " (Planning to build another)");
	} else if (unit->plan->nextbuild != NONUTYPE) {
	    sprintf(tmpbuf, " (Planning to build %s next)",
		    u_type_name(unit->plan->nextbuild));
	} else {
	    sprintf(tmpbuf, " (No more builds planned)");
	}
	strcat(spbuf, tmpbuf);
	draw_info_text(side, map, 40, 1, -1, spbuf);
    } else {
	draw_info_text(side, map, 40, 1, 50, "");
    }
#endif
    sprintf(spbuf, "");
    /* make this a "stacked_at" macro? */
    if ((mainunit = unit_at(unit->x, unit->y)) != NULL
	&& mainunit->nexthere != NULL) {
	strcpy(spbuf, "Others here  ");
	for_all_unit_types(i) nums[i] = 0;
	for_all_stack(unit->x, unit->y, occ) ++nums[occ->type];
	for_all_unit_types(i) {
	    if (nums[i] > 0) {
		sprintf(tmpbuf, "%d %1s  ", nums[i], utype_name_n(i, 1));
		strcat(spbuf, tmpbuf);
	    }
	}
    }
    draw_info_text(side, map, 40, 2, -1, spbuf);
    /* Describe the state of all the supplies. */
    sprintf(spbuf, "");
    for_all_material_types(r) {
	if (um_storage(u, r) > 0) {
	    sprintf(tmpbuf, "%s %d/%d  ",
		    m_type_name(r), unit->supply[r], um_storage(u, r));
	    strcat(spbuf, tmpbuf);
	}
    }
    draw_info_text(side, map, 40, 3, -1, spbuf);
    /* Describe the current plans, tasks, etc. */
    if (unit->plan) {
	char *pd;

	pd = plan_desig(unit->plan);
	draw_info_text(side, map, 0, 4, 40, pd);
    }
}

/* Display improvement can be achieved by padding out lines with blanks, */
/* then the lines need not be cleared before redrawing. */

draw_info_text(side, map, x, y, len, buf)
Side *side;
Map *map;
int x, y, len;
char *buf;
{
    int i;
    char full[BUFSIZE];

    if (len < 0) len = BUFSIZE;
    strcpy(full, buf);
    for (i = strlen(buf); i < len; ++i) full[i] = ' ';
    full[len-1] = '\0';
    draw_fg_text(side, map->infowin,
		 margin + x * side->ui->fw, y * side->ui->fh, full);
}

#if 0
		sprintf(tmpbuf, "%s", feature_name_at(map->curx, map->cury));
		if (temperatures_defined()) {
		    sprintf(tmpbuf+strlen(tmpbuf), "(T %d)",
			    temperature_at(map->curx, map->cury));
		}
		if (elevations_defined()) {
		    sprintf(tmpbuf+strlen(tmpbuf), "(El %d)",
			    elev_at(map->curx, map->cury));
		}
#endif

#if 0
/* Don't mix up the side you're looking at and the side you're displaying on */

draw_side_closeup(side, sidecloseup, side2)
Side *side, *side2;
SideCloseup *sidecloseup;
{
    int line = 0, u;

    sprintf(spbuf, "Side named \"%s\"", side2->name);
    if (side_lost(side2)) {
	sprintf(spbuf+strlen(spbuf), " (lost)");
    }
    draw_fg_text(side, sidecloseup->closeupwin,
		 margin, (line++) * side->ui->fh, spbuf);
/*    sprintf(spbuf, "Strategy is under %s control.",
	    (side->strategy ? "machine" : "no"));  */
    draw_fg_text(side, sidecloseup->closeupwin,
		 margin, (line++) * side->ui->fh, spbuf);
    sprintf(spbuf, "Ask Side %d, Time Out %d Seconds",
	    side2->doctrine->unitseveraskside, side2->doctrine->realtimeout);
    draw_fg_text(side, sidecloseup->closeupwin,
		 margin, (line++) * side->ui->fh, spbuf);
    for_all_unit_types(u) {
	sprintf(spbuf, "%s: %s ask side, %s bad terrain,",
		u_type_name(u),
		(u_doctrine(side2, u, everaskside) ? "may" : "never"),
		(u_doctrine(side2, u, avoidbadterrain) ? "avoid" : "ignore")
		);
	draw_fg_text(side, sidecloseup->closeupwin,
		     margin, (line++) * side->ui->fh, spbuf);
	sprintf(spbuf, "    build %s.",
		(u_doctrine(side2, u, buildalways) ? "always" : "sometimes"));
	draw_fg_text(side, sidecloseup->closeupwin,
		     margin, (line++) * side->ui->fh, spbuf);
    }
}
#endif

show_bool_popup(side)
Side *side;
{
#if 0
    int offset = margin;

    XClearWindow(side->ui->dpy, side->ui->boolpopup);
    draw_fg_text(side, side->ui->boolpopup,
		 offset, 0, side->ui->boolprompt);
    /* eventually make some buttons */
#endif
}

/* The prompt window consists of exactly one line, but we have to keep track */
/* of where the blank space begins, for when people type into it. */

show_string_popup(side)
Side *side;
{
#if 0
    int offset = margin;

    XClearWindow(side->ui->dpy, side->ui->stringpopup);
    draw_fg_text(side, side->ui->stringpopup,
		 offset, 0, side->ui->stringprompt);
    offset += string_width(side->ui->stringprompt,side->ui->textfont) + 10;
    draw_fg_text(side, side->ui->stringpopup,
		 offset, 0, side->ui->stringanswer);
    offset += string_width(side->ui->stringanswer,side->ui->textfont);
    draw_fg_text(side, side->ui->stringpopup, offset, 0, "_");
#endif
}

show_game_state(side, map)
Side *side;
Map *map;
{
    extern int paused;

    if (map == NULL) return;
    XClearWindow(side->ui->dpy, map->game);
    /* First line of the game state. */
    sprintf(spbuf, "%s", absolute_date_string(curturn));
    /* (should maybe provide an end date here too?) */
    if (world.yearlength > 1) {
	sprintf(tmpbuf, " (%s)", season_name(curturn));
	strcat(spbuf, tmpbuf);
    }
    draw_fg_text(side, map->game, margin, 0, spbuf);
    /* Second line of the game state. */
    sprintf(spbuf, "%s%s",
	    (paused ? "Paused" : ""),
	    "");
    draw_fg_text(side, map->game, margin, side->ui->fh, spbuf);
    show_global_clock(side, map);
}

/* The kernel calls this to update the global game state. */

update_turn_display(side, rightnow)
Side *side;
int rightnow;
{
    Map *map;

    if (active_display(side)) {
	for_all_maps(side, map) {
	    show_game_state(side, map);
	}
	if (rightnow) flush_output(side);
    }
}

update_action_display(side, rightnow)
Side *side;
int rightnow;
{
    Map *map;

    if (active_display(side)) {
	/* show state of actions */
	if (rightnow) flush_output(side);
    }
}

/* This displays the gamewide clock if defined. */
/* To get the traditional digital clock display, we need fixed-format digit
   printing. */

/* (should this be a mini-window inside global game window of a map?) */

show_global_clock(side, map)
Side *side;
Map *map;
{
#if 0
    int time = 0;
    int hour, minute, second;

/*    if (!period_giventime()) return;  */
    hour = time / 3600;  minute = (time / 60) % 60;  second = time % 60;
    if (active_display(side)) {
	XClearWindow(side->ui->dpy, map->clock);
	sprintf(spbuf, "%s%d:%s%d:%s%d",
		(hour < 10 ? "0" : ""), hour,
		(minute < 10 ? "0" : ""), minute,
		(second < 10 ? "0" : ""), second);
	draw_fg_text(side, side->ui->clock, 0, 0, spbuf);
    }
#endif
}

/* Updates to clock need to be sure that display changes immediately. */

update_clock_display(side, rightnow)
Side *side;
int rightnow;
{
    Map *map;

    if (active_display(side)) {
	for_all_maps(side, map) {
	    show_global_clock(side, map);
	}
	if (rightnow) flush_output(side);
    }
}

/* Write onto the list of sides. */

show_sides(side, map)
Side *side;
Map *map;
{
    Side *side2;
    int sy;

    if (map == NULL) return;
    map->sidespacing = side->ui->fh + 12;
    for_all_sides(side2) {
	draw_side_info(side, map, side2);
    }
}

/* Show info about a single side to some other side. */

draw_side_info(side, map, side2)
Side *side, *side2;
Map *map;
{
    int sx, sy;

    if (map == NULL || side2 == NULL) return;
    sx = margin + 16 + margin;
    sy = (side_number(side2) - 1) * map->sidespacing;
    draw_side_emblem(side, map->sides, 2, sy + 4, 8, 8, side2);
    if (side2->player) {
	sprintf(tmpbuf, "(%s)", player_desig(side2->player));
    } else {
	strcpy(tmpbuf, "");
    }
    sprintf(spbuf, "%s%s%s",
	    (side2->designer ? "(designer) " : ""),
	    short_side_title(side2),
	    tmpbuf);
    draw_fg_text(side, map->sides, sx, sy, spbuf);
    if (side_won(side2)) {
	/* what to do here? */
    } else if (side_lost(side2)) {
	/* Draw the line of ignominy through sides that have lost. */
	XSetForeground(side->ui->dpy, side->ui->gc, side->ui->fgcolor);
	XDrawLine(side->ui->dpy, map->sides, side->ui->gc,
		  0, sy + side->ui->fh / 2,
		  500, sy + side->ui->fh / 2);
    } else if (side2->ingame) {
	draw_side_progress(side, map, side2);
    } else {
	/* what to do here? */
    }
}

/* Display how far along the side is with moving its stuff. */
/* (needs some visual tuning) */

draw_side_progress(side, map, side2)
Side *side, *side2;
Map *map;
{
    int sx, sy, totacp, left = 0;

    if (map == NULL || side2 == NULL) return;
    sx = margin + 16 + margin;
    sy = (side_number(side2) - 1) * map->sidespacing;
    XSetClipMask(side->ui->dpy, side->ui->gc, None);
    XSetForeground(side->ui->dpy, side->ui->gc, side->ui->fgcolor);
    XDrawRectangle(side->ui->dpy, map->sides, side->ui->gc,
		   sx, sy + side->ui->fh, 100, 10);
    if (side2->finishedturn) {
	left = 1;
    } else {
	totacp = side_initacp(side2);
	if (totacp > 0) {
	    left = (100 * side_acp(side2)) / totacp;
	}
    }
    if (left > 0) {
	XFillRectangle(side->ui->dpy, map->sides, side->ui->gc,
		       sx + 1, sy + side->ui->fh + 1, left, 8);
    }
}

/* Fix only the list displays relevant to the given unit. */

update_unit_lists(side, unit)
Side *side;
Unit *unit;
{
#if 0
    List *list;

    if (active_display(side)) {
	for_all_lists(side, list) {
	    if (list->type == 1) {
		update_unit_list(side, list, unit);
	    }
	}
    }
#endif
}

/* Fix only the list displays relevant to the given type. */

update_unit_type_lists(side, u)
Side *side;
int u;
{
#if 0
    List *list;

    if (active_display(side)) {
	for_all_lists(side, list) {
	    if (list->type == 0) {
		update_unit_type_list(side, list, u);
	    }
	}
    }
#endif
}

/* General list display routine. */

show_list(side, map)
Side *side;
Map *map;
{
    XClearWindow(side->ui->dpy, map->list);
#if 0
    switch (list->type) {
      case 0:
	draw_unit_type_list(side, list);
	break;
      case 1:
	draw_unit_list(side, list);
	break;
    }
#endif
}

#if 0
/* Draw the entire list of unit types. */

/* (Need a way to draw only the ones present.) */

draw_unit_type_list(side, list)
Side *side;
List *list;
{
#if 0
    int u, pos = 2 * side->ui->fh, spacing = max(hhs[4], side->ui->fh);
    int power = 4;

    /* Draw a header on the list display. */
    draw_fg_text(side, list->listwin, 0, 0,
		 "All Unit Types");
    /* should do a dividing line */
    draw_fg_text(side, list->listwin,
		 hws[power] + margin, side->ui->fh,
		 " Num(Bld)     Tot   Lost");
    for_all_unit_types(u) {
	draw_hex_icon(side, list->listwin,
		      margin, pos, 0, power, side->ui->bgcolor);
	draw_unit_icon(side, list->listwin,
		       margin + (hws[power] - uws[power]) / 2,
		       pos + (hhs[power] - uhs[power]) / 2,
		       u, power, side->ui->fgcolor, side->ui->bgcolor);
	update_unit_type_list(side, list, u);
	pos += spacing;
    }
#endif
}

/* Alter the numbers for a single type of unit.  Should be called right
   after any changes.  Formatted to look nice, but kind of messy to set
   up correctly; display should not jump back and forth as the numbers
   change in size. */

update_unit_type_list(side, list, u)
Side *side;
List *list;
int u;
{
#if 0
    int xpos, ypos;

    if (active_display(side)) {
	/* Tack all the bits of info onto the end of the buffer. */
	sprintf(spbuf, "");
	/* Our unit total (right-justified) */
	if (side->units[u] > 0)	{
	    sprintf(tmpbuf, "%4d", side->units[u]);
	} else {
	    sprintf(tmpbuf, "    ");
	}
	strcat(spbuf, tmpbuf);
	/* Our units under construction (left-justified) */
	if (side->unitsbuilding[u] > 0) {
	    sprintf(tmpbuf, "(%d)", side->unitsbuilding[u]);
	} else {
	    sprintf(tmpbuf, "    ");
	}
	strcat(spbuf, tmpbuf);
	/* Our total possessions over the game */
	/* Make sure that subsequent write goes into valid string area. */
	strcat(spbuf,  "            ");
	if (total_gain(side, u) > 0) {
	    sprintf(tmpbuf, "%4d", total_gain(side, u));
	    sprintf(spbuf+10, "%s", tmpbuf);
	}
	/* Our total losses over the game */
	strcat(spbuf,  "            ");
	if (total_loss(side, u) > 0) {
	    sprintf(tmpbuf, "- %d", total_loss(side, u));
	    sprintf(spbuf+15, "%s", tmpbuf);
	}
	xpos = hws[4] + margin;
	ypos = 2 * side->ui->fh
	  + (u) * max(hhs[4], side->ui->fh)
	  + (hhs[4] - side->ui->fh) / 2;
	draw_fg_text(side, list->listwin, xpos, ypos, spbuf);
    }
#endif
}

/* Would be faster to stash these, but enough difference to care? */

total_gain(side, u)
Side *side;
int u;
{
    int i, total = 0;

/*    for (i = 0; i < DUMMYREAS; ++i)
      total += side_balance(side, u, i);  */
    return total;
}

total_loss(side, u)
Side *side;
int u;
{
    int i, total = 0;

/*    for (i = DUMMYREAS+1; i < NUMREASONS; ++i)
      total += side_balance(side, u, i);  */
    return total;
}

/* Unit lists can display all sorts of details. */

draw_unit_list(side, list)
Side *side;
List *list;
{
#if 0
    int i = 0, line = 0;
    Unit *unit;
    Side *loopside;

    if (0) {
	draw_fg_text(side, list->listwin, 1, (line++) * side->ui->fh,
		     "Units, in Normal Order");
	for_all_side_units(side, unit) {
	    if (alive(unit)) {
		sprintf(spbuf, "%-30s   HP %d/%d",
			unit_handle(side, unit), unit->hp, u_hp(unit->type));
	    } else {
		sprintf(spbuf, "---");
	    }
	    draw_fg_text(side, list->listwin, 1, (line++) * side->ui->fh,
			 spbuf);
	}
    } else if (1) {
	draw_fg_text(side, list->listwin, 1, (line++) * side->ui->fh,
		     "Units, in Acting Order");
	for (unit = side->actorder; unit != NULL; unit = unit->nextactor) {
	    if (alive(unit)) {
		sprintf(spbuf, "%-30s   HP %d/%d",
			unit_handle(side, unit), unit->hp, u_hp(unit->type));
	    } else {
		sprintf(spbuf, "---");
	    }
	    draw_fg_text(side, list->listwin, 1, (line++) * side->ui->fh,
			 spbuf);
	}
    } else {
	/* special-purpose lists? */
    }
#endif
}

/* Just change the part of the unit list relevant to the given unit. */

update_unit_list(side, list, unit)
Side *side;
List *list;
Unit *unit;
{
#if 0
    XClearWindow(side->ui->dpy, list->listwin);
    /* will be cleverer eventually */
    draw_unit_list(side, list, unit);
    flush_output(side);
#endif
}
#endif

/* General text drawer. */

draw_text(side, win, x, y, str, color)
Side *side;
Window win;
int x, y, color;
char *str;
{
    XFontStruct *font = side->ui->textfont;
    GC gc = side->ui->textgc;

    y += font->max_bounds.ascent;
    XSetForeground(side->ui->dpy, gc, color);
    /* If "color" is bg, assume we want reverse video. */
    if (color == side->ui->bgcolor) {
	XSetBackground(side->ui->dpy, gc, side->ui->fgcolor);
    } else {
	XSetBackground(side->ui->dpy, gc, side->ui->bgcolor);
    }
    XDrawImageString(side->ui->dpy, win, gc, x, y, str, strlen(str));
#ifdef STUPIDFLUSH
    XFlush(side->ui->dpy);
#endif STUPIDFLUSH
}

/* Frequently-called routine to draw text in the foreground color. */

draw_fg_text(side, win, x, y, str)
Side *side;
Window win;
int x, y;
char *str;
{
    draw_text(side, win, x, y, str, side->ui->fgcolor);
}

show_help(side)
Side *side;
{
#if 0
    int i, x = 0, y = 0, linestart = 0;
    int win = side->ui->help;
    char *str = side->ui->helpstring;

    clear();
    if (cur_at(side, win, x, y)) {
	for (i = 0; i < strlen(str); ++i) {
	    if (str[i] == '\n') {
		if (y < side->ui->wh[win]) {
		    cur_at(side, win, x, ++y);
		    linestart = i;
		} else {
		    cur_at(side, win, x, y-1);
		    standout();
		    addstr("--More--");
		    standend();
		    /* and record spot? */
		}
	    } else if (x + (i - linestart) < side->ui->ww[win]) {
		addch(str[i]);
	    }
	}
    }
#endif
}
