/* More stuff for the map widget. */

#include "conq.h"
#include "xtconq.h"

#include <X11/IntrinsicP.h>
/* #include <X11/StringDefs.h> */
#include "xtmapP.h"

#define viewdata int
#define side_view(s,x,y) 0

extern int bwid[];

/* The main map-drawing routine. */

/* static */ void
redisplay_map(w, evt, region)
Widget w;
XEvent *evt;
Region region;
{
    int y1, y2, y, x1, x2, adj;
    MapWidget mw = (MapWidget) w;
    Side *side;
    unsigned long mask;
    XGCValues values;

    side = mw->map.side;
    if (!active_display(side)) return;

    /* a hack */
    if (mw->map.gc == NULL) {
	mask = GCForeground | GCBackground;
	values.foreground = mw->map.fgcolor;
	values.background = mw->core.background_pixel;
	/* set linewidth etc here too? */
	mw->map.gc = XCreateGC(XtDisplay(mw), XtWindow(mw), mask, &values);
    }

    Dprintf("Displaying %sview at %d,%d, size %d,%d, center %d,%d\n",
	    (mw->map.isworld ? "world " : ""),
	    mw->map.vx, mw->map.vy, mw->map.vw, mw->map.vh,
	    mw->map.vcx, mw->map.vcy);
    /* Compute top and bottom rows to be displayed. */
    /* Include rows that will only be partly drawn. */
    y1 = mw->map.vy + mw->map.vh - 1;
    y2 = mw->map.vy;
    for (y = y1; y >= y2; --y) {
	/* Adjust the right and left bounds to fill the viewport as
	   much as possible, without going too far (the drawing code
	   will clip, but clipped drawing is still expensive). */
	/* could test by drawing viewport rect as lines... */
	adj = (y - mw->map.vy) / 2;
	/* If the world doesn't wrap, then we might have to stop
	   drawing before we reach the edge of the viewport. */
	/* (is this really reliable?) */
	if (area.xwrap) {
	    x1 = mw->map.vx - adj - 1;
	    x2 = mw->map.vx + mw->map.vw - adj + 1;
	} else {
	    int halfheight = area.height / 2;
	    
	    x1 = mw->map.vx - adj - 1;
	    x2 = mw->map.vx + mw->map.vw - adj + 1;  
	    x1 = max(0, min(x1, area.width-1)); /* use limitx instead? */
	    x2 = max(0, min(x2, area.width));
	    if (x1 + y > area.width + halfheight)
	      continue; /* x1 = area.width + halfheight - y;  */
	    if (x2 + y > area.width + halfheight)
	      x2 = area.width + halfheight - y;
	    if (x1 + y < 1 * halfheight)
	      x1 = 1 * halfheight - y;
	    if (x2 + y < 1 * halfheight)
	      continue; /* x2 = 1 * halfheight - 1 - y;  */
	    
	}
	draw_row(side, mw, x1, y, x2 - x1, FALSE);
	/* (should test for input events here, would respond better) */
    }
}

/* Put the point x, y in the center of the map. */

void
recenter(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
#if 0
    if ((x == mw->map.vcx) && (y == mw->map.vcy)) {
	/* no movement needed */
	/* can we assume the state of the display is valid? */
	/* could use oldx,y to fix... */
    } else {
	mw->map.vcx = x;  mw->map.vcy = y;  /* tell the widget */
/*	set_map_viewport(map);  */
/*	show_map(side, map, TRUE, TRUE);  */
    }
#endif
}

/* Ensure that given location is visible.  We also flush the input because
   any input relating to a different screen is probably worthless. */

/* should this happen to all maps or just the front one? */

put_on_screen(side, x, y)
Side *side;
int x, y;
{

    /* Ugly hack to prevent extra boxes being drawn during init - don't ask!*/
    if (x == 0 && y == 0) return;
    if (active_display(side)) {
	if (!in_middle(side, x, y)) {
	    recenter(side, side->ui->front, x, y);
	}
    }
}

/* Decide whether given location is not too close to edge of screen. */
/* We do this because it's a pain to move units when half the adjacent */
/* places aren't even visible.  This routine effectively places a lower */
/* limit of 5x5 for the map window. (I think) */

in_middle(side, x, y)
Side *side;
int x, y;
{
    Map *map = side->ui->front;

    int vcx = map->vcx, vw2 = map->vw2, odd = map->vw % 2;
    int vcy = map->vcy, vh2 = map->vh2;
    int left_hex, right_hex;
    
    /* changed so that units on top and bottom lines do the right thing. */
    /* this still isn't quite right! */
    if (!between(vcy-vh2+2, y, vcy+vh2-2) && between(1, y, area.height-2))
	return FALSE;
    left_hex = vcx - vw2 - ((y - vcy) / 2) - (y > vcy) + 3;
    right_hex = vcx + vw2 + odd - ((y - vcy) / 2) + ((y < vcy) && odd)
      - ((y > vcy) && !odd) - 3;

    if (left_hex < 0)
      return !between(right_hex, x, left_hex + area.width);
    else if (right_hex >= area.width)
      return !between(right_hex - area.width, x, left_hex);
    return between(left_hex, x, right_hex);
}

/* Decide if the position is visible in the front map. */
/* (should be "on some visible map"?) */

in_view(side, x, y)
Side *side;
int x, y;
{
    Map *map = side->ui->front;

    int vx = map->vx, vy = map->vy;
    int vw = map->vw, vh = map->vh;
    int vcx = map->vcx, vw2 = map->vw2, odd = map->vw % 2;
    int vcy = map->vcy, vh2 = map->vh2;
    int left_hex, right_hex;
    
    /* changed so that units on top and bottom lines do the right thing. */
    /* this still isn't quite right! */
    if (!between(vy, y, vy+vh) && between(0, y, area.height-1))
	return FALSE;
    left_hex = vx - ((y - vy) / 2) - 1;
    right_hex = vx + vw - ((y - vy) / 2) + 1;

    if (left_hex < 0)
      return !between(right_hex, x, left_hex + area.width);
    else if (right_hex >= area.width)
      return !between(right_hex - area.width, x, left_hex);
    return between(left_hex, x, right_hex);
}

/* Transform map coordinates into screen coordinates, relative to the given
   side.  Allow for cylindricalness and number of pixels in a hex. */

xform(side, mw, x, y, sxp, syp)
Side *side;
MapWidget mw;
int x, y, *sxp, *syp;
{
    /* this routine doesn't work on garbage points */
    if (!in_area(x, y)) {
	fprintf(stderr, "%d,%d shouldn't be xformed!\n", x, y);
    }
    /* Compute the pixel positions. */
    *sxp = ((mw->map.hw * (x - mw->map.vx))
	    + (mw->map.hw * (y - mw->map.vy)) / 2);
    *syp = mw->map.hch * ((mw->map.vy + mw->map.vh) - y - 1);
}

/* Un-transform screen coordinates (as supplied by mouse perhaps) into
   map coordinates.  Compute both a hex and a closest direction, just
   in case the caller needs to know.   This routine actually doesn't
   pay attention to hexagon boundaries, using only box-shaped areas,
   but in practice this seems to be OK. */

deform(side, mw, sx, sy, xp, yp, dirp)
Side *side;
MapWidget mw;
int sx, sy, *xp, *yp, *dirp;
{
    int adjust;

    /* Scale to hex coords then flip y. */
    *yp = (mw->map.vy + mw->map.vh) - (sy / mw->map.hch) - 1;
    /* Even/odd row adjustment gizmo (value in pixels). */
    adjust = (((*yp - mw->map.vy) & 1) ? (mw->map.hw / 2) : 0);
    /* Scale to hex then offset and correct for hex obliqueness. */
    *xp = mw->map.vx + ((sx - adjust) / mw->map.hw) - (*yp - mw->map.vy) / 2;
    /* Wrap coords as usual. */
    if (area.xwrap) *xp = wrapx(*xp);
    /* Compute the closest edge, as a direction.  Note that this doesn't
       return all dirs (since would return adj hex pos with opposite dir -
       valid because also refers to the same edge). */
    if (((sx - adjust) % mw->map.hw) > mw->map.hw/2) {
	*dirp = ((sy % mw->map.hch < (mw->map.hh - mw->map.hch)) ? NE : EAST);
    } else {
	*dirp = ((sy % mw->map.hch < (mw->map.hh - mw->map.hch)) ? NW : WEST);
    }
    Dprintf("Pixel %d,%d (relative %d,%d) -> hex %d,%d dir %d\n",
	    sx, sy, (sx - adjust) % mw->map.hw, sy % mw->map.hch, *xp, *yp, *dirp); 
}

/* Draw an individual detailed hex, as a row of one, on all maps. */

draw_hex(side, x, y, flushit)
Side *side;
int x, y;
int flushit;
{
#if 0
    int sx, sy;
    Map *map;

    if (active_display(side)) {
	for_all_maps(side, map) {
	    draw_row(side, map, x, y, 1, TRUE);
	}
	if (flushit) flush_output(side);
    }
#endif
}

update_cell_display(side, x, y, flushit)
Side *side;
int x, y;
int flushit;
{
#if 0
    int sx, sy;
    Map *map;

    if (active_display(side)) {
	for_all_maps(side, map) {
	    draw_row(side, map, x, y, 1, TRUE);
	}
	if (flushit) flush_output(side);
    }
#endif
}

/* Return the *predominant* color within the given hex. */

long
hex_color_at(side, w, x, y)
Side *side;
Widget w;
int x, y;
{
    MapWidget mw = (MapWidget) w;

    if (mw->map.drawterrain && side_view(side, wrap(x), y) != UNSEEN) {
	/* account for snow? */
	return mw->map.hexcolor[terrain_at(wrap(x), y)];
    } else {
	return mw->core.background_pixel;
    }
}

/* The basic map drawing routine does an entire row at a time, which yields
   order-of-magnitude speedups (!).  Actual nature of speedup is device
   specific, this routine concentrates on getting all the different
   display elements on the screen. */

draw_row(side, mw, x0, y0, len, clearit)
Side *side;
MapWidget mw;
int x0, y0, len, clearit;
{
    int empty = TRUE;
    char ch;
    int i = 0, x, sx, sy;

    if (len >= 1000) run_error("Display too wide!");
#if 0
    if (0 /* clearit */) {
	xform(side, map, x0, y0, &sx, &sy);
	{
	    int i;
	    char hexbuf[1000];

	    for (i = 0; i < len; ++i) {
		hexbuf[i] = 255; /* ? */
	    }
	    /* Draw a collection of bg-colored hexes. */
	    draw_uniform_segment(mw, sx, sy, hexbuf, len,
				 /* need a generic fill "terrain type" */
				 sd(side)->usewhat[map->power][0],
				 side->ui->bgcolor);
	}
    }
#endif
    if (mw->map.drawterrain) {
	draw_terrain_row(side, mw, x0, y0, len);
    }
    /* Draw sparse things on top of the basic row. */
    for (x = x0; x < x0 + len; ++x) {
	/* The relative ordering of these is quite important.  Note that
	   each should be prepared to run independently also, since the
	   other displays might have been turned off. */
#ifdef LINEAR
	if (borders_defined() && mw->map.drawborders && mw->map.drawterrain)
	  draw_borders(side, mw, x, y0);
	/* This isn't in the right place to draw roads underneath */
	/* the units of the previous row... */
	if (connections_defined() && mw->map.drawconns && mw->map.drawterrain)
	  draw_connections(side, mw, x, y0);
#endif /* LINEAR */
#ifdef PEOPLE
	if (people_defined() && mw->map.drawpeople) {
	    draw_people(side, mw, x, y0);
	}
#endif /* PEOPLE */
	if (mw->map.drawunits) {
	    draw_units(side, mw, x, y0);
	}
    }
    /* Need a second loop so "hex-crossing" bits don't get half-destroyed. */
    for (x = x0; x < x0 + len; ++x) {
	/* (should be global to map drawing somehow?) */
	if (mw->map.drawnames) {
	    draw_legend(side, mw, x, y0);
	}
	/* This should be the "topmost" thing. */
	if (mw->map.curunit && x == mw->map.curunit->x && y0 == mw->map.curunit->y) {
	    draw_cursor(side, mw);
	}
    }
}

/* Return the color of a terrain icon overlaying a colored hex. */

terricon_color(side, x, y)
Side *side;
int x, y;
{
    return ((side_view(side, wrap(x), y) == UNSEEN) ? side->ui->bgcolor :
	     side->ui->bgcolor);
}

/* Draw a single unit icon as appropriate.  This *also* has a bunch of
   details to worry about: centering of icon in hex, clearing a rectangular
   area for the icon, picking a color for the unit, using either a bitmap
   or font char, and adding a side emblem.

   Must also be careful not to draw black-on-black for units in space. */

draw_units(side, mw, x, y)
Side *side;
MapWidget mw;
int x, y;
{
    viewdata view = side_view(side, wrap(x), y);
    int sx, sy, ucolor, hcolor, n, p = mw->map.power, u;
    int terr = terrain_at(wrap(x), y);
    Side *side2;
    
    if (view != UNSEEN && view != EMPTY) {
#ifdef LINEAR
	/* Fix up any connections or borders */
	/* do borders usually need fixup? */
	if (borders_defined() && mw->map.drawborders && mw->map.drawterrain) {
	    draw_borders(side, mw, x, y);
	}
	if (connections_defined() && mw->map.drawconns && mw->map.drawterrain) {
	    draw_connections(side, mw, x, y);
	}
#endif /* LINEAR */
	xform(side, mw, x, y, &sx, &sy);
	u = vtype(view);
	side2 = side_n(vside(view));
	if (side != side2) { /* or ally? or under direct observation? */
	    /* Just the view data to work from, can't report much. */
	    /* Determine the preferred color for the icon. */
	    /* (what about multi-colored icons?) */
#if 0
	    ucolor = (allied_side(side, side2) ? side->ui->owncolor :
		      (enemy_side(side, side2) ? side->ui->enemycolor :
		       side->ui->/* neutcolor */ owncolor));
#endif
	    ucolor = mw->map.fgcolor;
	    /* When does the underlying hex need to be redrawn? */
	    /* Determine the general background color here. */
	    hcolor = hex_color_at(side, mw, x, y);
	    /* Now splat down the unit icon proper, centering it in
	       the hex. */
	    /* (what if displaying multiple units here?) */
	    draw_unit_icon(side, mw,
			   sx + (mw->map.hw -
				 8 /* unit_icon_width(side, p, u) */) / 2,
			   sy + (mw->map.hh -
				 8 /* unit_icon_height(side, p, u) */) / 2,
			   u, p, ucolor, hcolor);
	    /* Draw the emblem of the unit's side, allowing omission
	       if the circumstances warrant. */
	    /* would be better to put emblem on unit's UR corner? */
	    /* even better to put emblem at icon-specified point */
	    draw_side_emblem(side, mw,
			     sx + mw->map.hw / 2,
			     sy + (mw->map.hw - mw->map.uw) / 2,
			     side2);
	} else {
	    /* One of our own, we get to look at the entire stack. */
	    int num = 0, nrows, sxoff, syoff, sdx, sdy;
	    Unit *unit;

	    for_all_stack(x, y, unit) ++num;
	    nrows = isqrt(num);
	    if (nrows * nrows < num) ++nrows;  /* round up */

	    /* Compute offset of unit icon layout from hex origin. */
	    sxoff = (mw->map.hw - (nrows * unit_icon_width(side, p, 0))) / 2;
	    syoff = (mw->map.hh - (nrows * unit_icon_height(side, p, 0))) / 2;
	    sdx = sdy = 0;
	    num = 0;
	    for_all_stack(x, y, unit) {
		u = unit->type;
		side2 = unit->side;
		/* Determine the preferred color for the icon. */
		/* (what about multi-colored icons?) */
#if 0
		ucolor = (allied_side(side, side2) ? side->ui->owncolor :
			  (enemy_side(side, side2) ? side->ui->enemycolor :
			   side->ui->/* neutcolor */ owncolor));
#endif
	    ucolor = mw->map.fgcolor;
		/* When does the underlying hex need to be redrawn? */
		/* Determine the general background color here. */
		hcolor = hex_color_at(side, mw, x, y);
		/* Now splat down the unit icon proper, centering it in
		   the hex. */
		/* (what if displaying multiple units here?) */
		draw_unit_icon(side, mw,
			       sx + sxoff + sdx, sy + syoff + sdy,
			       u, p, ucolor, hcolor);
		/* Draw the emblem of the unit's side, allowing omission
		   if the circumstances warrant. */
		/* would be better to put emblem on unit's UR corner? */
		/* even better to put emblem at icon-specified point */
		draw_side_emblem(side, mw,
				 sx + mw->map.hw / 2,
				 sy + (mw->map.hw - mw->map.uw) / 2,
				 side2);
		if ((++num % nrows) == 0) {
		  sdx = 0;  sdy += unit_icon_height(side, p, 0);
		} else {
		  sdx += unit_icon_width(side, p, 0);
		}
	    }
	}
    }
}

/* Splash a unit image (either bitmap or font char) onto some window. */

/* (Add support for full-color icons?) */

draw_unit_icon(side, mw, sx, sy, u, color, colorunder)
Side *side;
MapWidget mw;
int sx, sy, u, color, colorunder;
{
    char buf[1];
    int p = mw->map.power;
    int uw = 8 /* sd(side)->unitw[p][u] */, uh = 8 /* sd(side)->unith[p][u] */;
    int bgcolor;
    Display *dpy = XtDisplay(mw);
    Window win = XtWindow(mw);
    GC gc = mw->map.gc;

    /* If the icon would be hard or impossible to see by itself, and doesn't
       have a mask, then draw a box in a contrasting color underneath. */
    if (1 /* sd(side)->unitmasks[p][u] == None */
	&& 0 /* low_contrast(side, color, colorunder) */) {
	bgcolor = (0 /* low_contrast(side, color, side->ui->bgcolor) */ ?
		   mw->map.fgcolor : side->ui->bgcolor);
	if (color == bgcolor) init_warning("Color confusion");
        XSetForeground(dpy, gc, bgcolor);
	XSetClipMask(dpy, gc, None);
	XFillRectangle(dpy, win, gc, sx, sy, uw, uh);
    }
    if (0 /* sd(side)->unitfonts[p][u] != NULL */) {
	XSetForeground(dpy, gc, color);
	buf[0] = sd(side)->unitchars[p][u];
	sy += (sd(side)->unitfonts[p][u])->max_bounds.ascent;
	XDrawString(dpy, win, gc, sx, sy, buf, 1);
    } else if (0 /* sd(side)->unitpics[p][u] != None */) {
	XSetClipOrigin(dpy, gc, sx, sy);
	/* Draw the unit icon mask if we have one. */
	if (sd(side)->unitmasks[p][u] != None) {
	    XSetForeground(dpy, gc, mw->map.fgcolor);
	    XSetClipMask(dpy, gc,
			 sd(side)->unitmasks[p][u]);
	    XFillRectangle(dpy, win, gc, sx, sy, uw, uh);
	}
	/* Draw the unit icon proper. */
	XSetForeground(dpy, gc, color);
	XSetClipMask(dpy, gc,
		     sd(side)->unitpics[p][u]);
	XFillRectangle(dpy, win, gc, sx, sy, uw, uh);
    } else {
	/* Pretty desperate, or else we're at a scale that doesn't matter. */
	XSetClipMask(dpy, gc, None);
	/* Draw an edge, if the box will be more than a couple pixels. */
	if (uw >= 2) {
	    XSetForeground(dpy, gc, mw->map.bgcolor);
	    XFillRectangle(dpy, win, gc, sx, sy, uw, uh);
	}
	/* Draw a filled box inside the edge. */
	XSetForeground(dpy, gc, color);
	XFillRectangle(dpy, win, gc, sx + 1, sy + 1, uw - 2, uh - 2);
	/* Might be cool to draw unit type name or char inside/instead */
    }
}

#ifdef PEOPLE

/* Indicate what kind of people are living in the given hex. */

draw_people(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
#if 0
    viewdata view = side_view(side, wrap(x), y);
    int sx, sy;
    Side *side2;
    int pop, pcolor;
	
    pop = people_side_at(wrap(x), y);
    if (pop != NOBODY) {
	side2 = side_n(pop);
	/* need to work on rules for color of people emblem */
	pcolor = (allied_side(side, side2) ? side->ui->owncolor :
		  (enemy_side(side, side2) ? side->ui->enemycolor :
		   side->ui->neutcolor));
#if 0
	if (pcolor == side->ui->owncolor && 
	    (0  ||
	     side->ui->monochrome ||
	     (0 /* mw->map.showmode == TERRICONS */)))
	  pcolor = side->ui->fgcolor;
#endif
	xform(side, map, x, y, &sx, &sy);
	/* compute pcolor inside drawing routine? */
	/* should this do a mask or bg also? */
	draw_people_emblem(side, XtWindow(mw), sx, sy, side2);
    }
#endif
}

#endif /* PEOPLE */

#ifdef LINEAR

/* Draw three borders of the given hex. */

/* (do we need another routine to repair all six borders?) */

draw_borders(side, mw, x, y)
Side *side;
MapWidget mw;
int x, y;
{
    int sx, sy, b;

    if (bwid[mw->map.power] == 0) return;
    xform(side, mw, x, y, &sx, &sy);
    if ((b = border_at(x, y, WEST)) != NOBORDER
	&& seen_border(side, x, y, WEST)) {
	draw_border_line(side, mw, sx, sy, WEST, b);
    }
    if ((b = border_at(x, y, NW)) != NOBORDER
	&& seen_border(side, x, y, NW)) {
	draw_border_line(side, mw, sx, sy, NW,   b);
    }
    if ((b = border_at(x, y, NE)) != NOBORDER
	&& seen_border(side, x, y, NE)) {
	draw_border_line(side, mw, sx, sy, NE,   b);
    }
}

/* The border has been seen if the hex on either side has been seen. */

seen_border(side, x, y, dir)
Side *side;
int x, y, dir;
{
    return (side_view(side, x, y) != UNSEEN
	    || side_view(side, x + dirx[dir], y + diry[dir]) != UNSEEN);
}

/* Draw a single border on the given edge of the given hex. */

draw_border_line(side, mw, sx, sy, dir, b)
Side *side;
MapWidget mw;
int sx, sy, dir, b;
{
    int p = mw->map.power;

    if (bwid[p] == 0) return;
    /* (Can this be skipped if color same as adj hexes?) */
    XSetForeground(XtDisplay(mw), mw->map.gc, mw->map.hexcolor[b]);
    XSetClipMask(XtDisplay(mw), mw->map.gc, None);
    XSetLineAttributes(XtDisplay(mw), mw->map.gc,
		       bwid[p], LineSolid, CapButt, JoinMiter); 
    XDrawLine(XtDisplay(mw), XtWindow(mw), mw->map.gc,
	      sx + bsx[p][dir], sy + bsy[p][dir],
	      sx + bsx[p][dir+1], sy + bsy[p][dir+1]);
}

/* This draws six half-connections.  It also only draws the
   connection if the underlying terrain is different. */

draw_connections(side, mw, x, y)
Side *side;
MapWidget mw;
int x, y;
{
    int sx, sy, dir, b, t = terrain_at(x, y), p = mw->map.power;
    int cx = hws[p] / 2, cy = hhs[p] / 2;

    if (bwid[p] == 0 || side_view(side, x, y) == UNSEEN) return;
    xform(side, mw, x, y, &sx, &sy);
    for_all_directions(dir) {
	if ((b = connection_at(x, y, dir)) != NOCONNECTION && b != t) {
	    /* (can we cut down on calls to this? alloc special gc perhaps?) */
	    /* (or set only when map power is changed) */
	    XSetLineAttributes(XtDisplay(mw), mw->map.gc,
			       bwid[p], LineSolid, CapButt, JoinMiter); 
	    XSetForeground(XtDisplay(mw), mw->map.gc, mw->map.hexcolor[b]);
	    XSetClipMask(XtDisplay(mw), mw->map.gc, None);
	    XDrawLine(XtDisplay(mw), XtWindow(mw), mw->map.gc,
		      sx + cx, sy + cy,
		      sx + cx + lsx[p][dir], sy + cy + lsy[p][dir]);
	}
    }
}

#endif /* LINEAR */

/* Draw any text that should be associated with this hex. */

/* (could precompute what the string will lap over and move or truncate str),
   should be deterministic for each mag, so redraw doesn't scramble */

/* do geofeatures, label at a hex with nothing else, and declared as the
   feature's "center" */

/* Black/white text should be consistent for each period, use mask only
   to fix difficulties. */

draw_legend(side, mw, x, y)
Side *side;
MapWidget mw;
int x, y;
{
    int hasdark = FALSE, haslight = FALSE, maskit = FALSE;
    int xx, numhexes = 1, color, sx, sy;
    char legend[100];
    viewdata view = side_view(side, x, y);
    Unit *unit;

    if (mw->map.power < 3 || view == UNSEEN)  return;
    /* Draw a unit's name or number. */
    /* (This will only do top unit in hex, what about others?) */
    if (mw->map.drawunits
	&& (unit = unit_at(x, y)) != NULL) {
	/* still has a bug if types happen to match - need to use view date
	   instead? */
	if (view == EMPTY || vtype(view) != unit->type) return;
	if (unit->name != NULL) {
	    strcpy(legend, unit->name);
	} else if (unit->number > 0) {
	    sprintf(legend, "%d%s",
		    unit->number, ordinal_suffix(unit->number));
	} else {
	    /* Turns out we have nothing to make a legend with. */
	    return;
	}
	/* Look at all the hexes that the legend will overlap. */
	numhexes = (strlen(legend) * 8) / hws[mw->map.power];
	for (xx = x; xx < x + numhexes; ++xx) {
	    color = hex_color_at(side, mw, xx, y);
#if 0
	    if (low_contrast(side, color, side->ui->fgcolor)) {
		haslight = TRUE;
	    }
	    if (low_contrast(side, color, side->ui->bgcolor)) {
		hasdark = TRUE;
	    }
#endif
	    /* (should also account for borders and connections?) */
	    /* (should also see which other hexes have legends, decide
	       how to move this one) */
	}  
	/* some periods might be better with all white text */
	color = mw->core.background_pixel;
	if (hasdark) maskit = TRUE;
	/* (should also be done for any "noisy" background) */
/*	if (side->ui->monochrome) maskit = TRUE;  */
	xform(side, mw, x, y, &sx, &sy);
	draw_legend_text(side, mw,
			 /* or derive from unit's loc within hex... */
			 sx + mw->map.hw/2 + mw->map.uw/2 - 1,
			 sy + mw->map.hh/2 - mw->map.uh/2 + 1,
			 legend, color, maskit);
    } else {
	/* feature object should specify legend's position */
	extern GeoFeature *feature_at();
	GeoFeature *feature;

	feature = feature_at(x, y);
	if (feature) {
	    if (feature->size == 1) {
		int pixlen;

		strcpy(legend, feature_name_at(x, y));
		pixlen = strlen(legend) * 8;
		xform(side, mw, x, y, &sx, &sy);
		draw_legend_text(side, mw,
				 sx + 1 + (pixlen > mw->map.hw ? 2 :
					   (mw->map.hw/2 - pixlen / 2)),
				 sy - 6 + mw->map.hh / 2,
				 legend, side->ui->fgcolor, TRUE);
	    }
	}
    }
}

/* Map legends are stencils usually. */

draw_legend_text(side, mw, sx, sy, str, color, maskit)
Side *side;
MapWidget mw;
int sx, sy, color, maskit;
char *str;
{
#if 0
    XSetFont(sdd(side), sd(side)->textgc,
	     (sd(side)->ulegendfonts[power][0])->fid);
    sy += (sd(side)->ulegendfonts[power][0])->max_bounds.ascent;
#endif
    XSetForeground(XtDisplay(mw), mw->map.gc, mw->map.fgcolor);
    XSetBackground(XtDisplay(mw), mw->map.gc, mw->core.background_pixel);
/*    XSetForeground(XtDisplay(mw), mw->map.gc, color); */
    if (1 /* maskit */) {
      /* should be bgcolor if color == fgcolor */
/*	XSetBackground(sdd(side), mw->map.gc, side->ui->fgcolor);  */
	XDrawImageString(sdd(side), XtWindow(mw), mw->map.gc,
			 sx, sy, str, strlen(str));
    } else {
	XDrawString(XtDisplay(mw), XtWindow(mw), mw->map.gc,
		    sx, sy, str, strlen(str));
    }
}


draw_submap_boxes(side, worldmap)
Side *side;
Map *worldmap;
{
#if 0
    Map *map;

    for_all_maps(side, map) {
	if (!mw->map.isworld) {
	    draw_box(side, map, worldmap);
	}
    }
#endif
}

draw_box(side, map, worldmap)
Side *side;
Map *map, *worldmap;
{
#if 0
    int mm2, sx1, sy1, sx2, sy2;

    if (map && !mw->map.isworld && worldmap && worldmw->map.isworld) {
	mm2 = hws[worldmw->map.power] / 2;
	xform(side, worldmap, mw->map.vx, mw->map.vy, &sx1, &sy1);
	xform(side, worldmap,
	      mw->map.vx + mw->map.vw - mw->map.vh / 2, mw->map.vy + mw->map.vh, &sx2, &sy2);
	sx1 += mm2;  sy1 -= mm2;  sx2 += mm2;  sy2 += mm2;
	invert_box(side, worldXtWindow(mw), sx1, sy1, sx2, sy2, mm2);
	/* Draw a second box if we're wrapped around. */
/*	invert_box(side, worldmap, mw->map.vcx - area.width, mw->map.vcy);
	mw->map.lastvcx = mw->map.vcx;  mw->map.lastvcy = mw->map.vcy;
*/
    }
#endif
}

undraw_box(side, map, worldmap)
Side *side;
Map *map, *worldmap;
{
#if 0
    int mm2, sx1, sy1, sx2, sy2;

    if (map && !mw->map.isworld && worldmap && worldmw->map.isworld) {
	mm2 = hws[worldmw->map.power] / 2;
	xform(side, worldmap, mw->map.vx, mw->map.vy, &sx1, &sy1);
	xform(side, worldmap,
	      mw->map.vx + mw->map.vw, mw->map.vy + mw->map.vh, &sx2, &sy2);
	sx1 += mm2;  sy1 -= mm2;  sx2 += mm2;  sy2 += mm2;
	invert_box(side, worldXtWindow(mw), sx1, sy1, sx2, sy2, mm2);
    }
#endif
}

/* Cursor drawing also draws the unit in some other color if it's not the
   "top-level" unit in a hex. */

/* "color unders" here are not correct */

draw_cursor(side, map)
Side *side;
Map *map;
{
#if 0
    int sx, sy;

    if (active_display(side)
	&& mw->map.curunit
	&& alive(mw->map.curunit)
	&& mw->map.curunit->x >= 0) {
	xform(side, map, mw->map.curunit->x, mw->map.curunit->y, &sx, &sy);
	/* Maybe highlight the unit that the cursor is showing. */
	if (mw->map.curunit != NULL && mw->map.curunit->transport != NULL) {
	    if (0 /* side->ui->monochrome */) {
#if 0
		draw_hex_icon(side, XtWindow(mw),
			      sx, sy,
			      terrain_at(mw->map.curunit->x, mw->map.curunit->y),
			      mw->map.power, side->ui->bgcolor);
#endif
		draw_unit_icon(side, XtWindow(mw),
			       sx + (mw->map.hw - mw->map.uw) / 2,
			       sy + (mw->map.hh - mw->map.uh) / 2,
			       mw->map.curunit->type, mw->map.power,
			       side->ui->fgcolor, side->ui->bgcolor);
	    } else {
		draw_unit_icon(side, XtWindow(mw),
			       sx + (mw->map.hw - mw->map.uw) / 2,
			       sy + (mw->map.hh - mw->map.uh) / 2,
			       mw->map.curunit->type, mw->map.power,
			       side->ui->diffcolor, side->ui->bgcolor);
	    }
	}
	/* Draw the cursor icon proper. */
	draw_cursor_icon(side, map, sx, sy);
    }
#endif
}

/* Get rid of cursor by redrawing the hex. */
/* (Overkill, should only do for particular map) */

erase_cursor(side, map)
Side *side;
Map *map;
{
#if 0
    if (mw->map.curunit) {
	int lastx = mw->map.curunit->x, lasty = mw->map.curunit->y;

	mw->map.curunit = NULL;
	draw_row(side, map, lastx, lasty, 1, TRUE);
    }
#endif
}

/* Draw a splat visible to both sides at a given location.  Several splats
   available, depending on the seriousness of the hit.  Make an extra-flashy
   display when The Bomb goes off.  Because of the time delays involved, we
   have to update both sides' displays more or less simultaneously.  */

draw_blast(unit, es, hit)
Unit *unit;
Side *es;
int hit;
{
#if 0
    int ux = unit->x, uy = unit->y, sx, sy, i, btype;
    Side *us = unit->side, *side;
    Map *map;

    if (0 /* hit >= period_nukehit() */) {
#if 0
	/* or all sides? */
	if (active_display(us)) invert_whole_map(us);
	if (active_display(es)) invert_whole_map(es);
	sleep(1);
	/* may need a time delay if X is too speedy */
	if (active_display(us)) invert_whole_map(us);
	if (active_display(es)) invert_whole_map(es);
	for (i = 0; i < 4; ++i) {
	    if (active_display(us)) draw_mushroom(us, ux, uy, i);
	    if (active_display(es)) draw_mushroom(es, ux, uy, i);
	    if (i != 2 && (active_display(us) || active_display(es))) sleep(1);
	}
	if (active_display(us) || active_display(es)) sleep(1);
#endif
    } else {
	btype = ((hit >= unit->hp) ? 2 : ((hit > 0) ? 1 : 0));
	for_all_sides(side) {
	    if (active_display(side) && cover(side, ux, uy) > 0) {
		for_all_maps(side, map) {
		    if (mw->map.power >= 4) { /* what other tests? */
			xform(side, map, ux, uy, &sx, &sy);
			/* should be able to draw different size blasts
			   for each map zoom */
			draw_blast_icon(side, XtWindow(mw),
					sx + (mw->map.hw / 2 - 10),
					sy + (mw->map.hh / 2 - 11),
					btype,
					(side == us ? side->ui->enemycolor :
					 (side == es ? side->ui->owncolor :
					  side->ui->neutcolor)));
			flush_output(side);
		    } else {
			/* just flicker the screen a little? */
		    }
		}
	    }
	}
    }
#endif
}

/* Invert the outline box on the world map.  This is a little tricky, */
/* because we want the lines to run through the middle of the world's */
/* hexes, and because the line drawn should not overlap (or the overlaps */
/* will be doubly inverted and look strange). */

invert_box(side, win, sx1, sy1, sx2, sy2, mm2)
Side *side;
Window win;
int sx1, sy1, sx2, sy2, mm2;
{
    if (active_display(side)) {
	XSetFunction(sdd(side), sd(side)->gc, GXinvert);
	/* is this next call really necessary? */
	XSetLineAttributes(sdd(side), sd(side)->gc,
			   1, LineSolid, CapButt, JoinMiter);
	/* do drawrect instead? */
	XDrawLine(sdd(side), win, sd(side)->gc,
		  sx1, sy1, sx2, sy1);
	XDrawLine(sdd(side), win, sd(side)->gc,
		  sx2, sy1-1, sx2, sy2+1);
	XDrawLine(sdd(side), win, sd(side)->gc,
		  sx2, sy2, sx1, sy2);
	XDrawLine(sdd(side), win, sd(side)->gc,
		  sx1, sy2+1, sx1, sy1-1);
	XSetFunction(sdd(side), sd(side)->gc, GXcopy);
    }
}

enum whattouse terrstyles[] =
  { useblocks, useblocks, useblocks,
    /* usepictures, usepictures, */ useblocks, usepolygons,
    usepolygons, usepolygons, usepolygons };

enum whattouse
hex_style_at(side, mw, x, y)
Side *side;
MapWidget mw;
int x, y;
{
    return (side_view(side, wrap(x), y) != UNSEEN ?
	    terrstyles[mw->map.power] :
	    dontdraw);
}

/* This interfaces higher-level drawing decisions to the rendition of
   individual pieces of display.  The rendering technique chosen depends
   on what the init code has decided is appropriate given what it found
   during init and what magnification the display is at.

   This routine is performance-critical;  any improvements will probably
   have a noticeable effect on the display. */

char hexbuf[1000];

draw_terrain_row(side, mw, x0, y0, len)
Side *side;
MapWidget mw;
int x0, y0, len;
{
    int empty;
    int x, x1, sx, sy, i = 0;
    long color, segcolor;
    enum whattouse style, segstyle;
    Display *dpy = XtDisplay(mw);
    Window win = XtWindow(mw);
    GC gc = mw->map.gc;
    int power = mw->map.power, dogrid = mw->map.drawhexedges, dofill = 1;
    int hw = hws[power], hc = hcs[power], hh = hhs[power];

    x1 = x0;
    segcolor = hex_color_at(side, mw, wrap(x0), y0);
    segstyle = hex_style_at(side, mw, wrap(x0), y0);
    for (x = x0; x <= x0 + len; ++x) {
	color = hex_color_at(side, mw, wrap(x), y0);
	style = hex_style_at(side, mw, wrap(x), y0);
	if (x == x0 + len
	    || color != segcolor
	    || style != segstyle
	    || segstyle == usepictures
	    || segstyle == usepolygons) {
	    /* don't draw solid bg color */
	    if (segstyle != dontdraw
		&& segcolor != mw->core.background_pixel) {
		xform(side, mw, x1, y0, &sx, &sy);
		XSetForeground(dpy, gc, segcolor);
		switch (segstyle) {
		  case useblocks:
		    /* could maybe adjust to go along middle of hex borders? */
		    XSetClipMask(dpy, gc, None);
		    /* need to use tile sometimes */
		    XFillRectangle(dpy, win, gc, sx, sy, i * hw, hc);
		    break;
		  case usepolygons:
		    XSetClipMask(dpy, gc, None);
		    draw_hex_polygon(dpy, win, gc, sx, sy, power, dogrid);
		    break;
		}
	    }
	    i = 0;
	    x1 = x;
	    segcolor = color;
	    segstyle = style;
	}
	++i;
    }
}

#if 0
		    /* should use tiling within big hexes. */
		    if (0 /* dofill && sd(side)->terrpics[power][buf[i]] != None */) {
#if 0
			XSetFillStyle(dpy, gc, FillTiled);
			XSetTile(dpy, gc,
				 sd(side)->terrpics[power][buf[i]]);
			XSetTSOrigin(dpy, gc, sx, sy);
			draw_hex_polygon(dpy, win, gc, sx, sy, power, dogrid);
			XSetFillStyle(dpy, gc, FillSolid);
#endif
		    } else {
#endif

#if 0
		  case usepictures:
		    if (dogrid) {
			icon = sd(side)->bhexpic;
			XSetClipMask(dpy, gc, icon);
		    } else if (!dofill) {
			icon = sd(side)->hexpic;
			XSetClipMask(dpy, gc, icon);
		    }
		    for (i = 0; i < len; ++i) {
			if (dofill) {
			    icon = sd(side)->terrpics[power][buf[i]];
			    XSetClipMask(dpy, gc, icon);
			}
			XSetClipOrigin(dpy, gc, sx, sy);
			XFillRectangle(dpy, win, gc,
				       sx, sy, hw, hh);
			/* Increment to the next element of the row. */
			sx += hw;
		    }
		    break;
		  case usefontchars:
		    for (i = 0; i < len; ++i) {
			/* Replace the terrain types with appropriate characters. */
			if (dofill) {
			    buf[i] = sd(side)->terrchars[power][buf[i]];
			} else if (dogrid) {
			    buf[i] = 'o';
			} else {
			    buf[i] = 'O';
			}
		    }
		    sy += sd(side)->iconfont->max_bounds.ascent;
		    XSetClipMask(dpy, gc, None);
		    XDrawString(dpy, win, gc, sx, sy, buf, len);
		    break;
#endif

/* The cursor icon highlights a unit of interest - should be called a
   "selection highlight" or something similar. */

draw_cursor_icon(side, map, sx, sy)
Side *side;
Map *map;
int sx, sy;
{
#if 0
    if (mw->map.power == 4) {
	XSetClipOrigin(sdd(side), sd(side)->varicongc, sx, sy);
	XSetForeground(sdd(side), sd(side)->varicongc, side->ui->bgcolor);
	XSetClipMask(sdd(side), sd(side)->varicongc, sd(side)->boxmask);
	XFillRectangle(sdd(side), XtWindow(mw), sd(side)->varicongc,
		       sx, sy, mw->map.hw, mw->map.hh);
	XSetForeground(sdd(side), sd(side)->varicongc, side->ui->fgcolor);
	XSetClipMask(sdd(side), sd(side)->varicongc, sd(side)->boxcurs);
	XFillRectangle(sdd(side), XtWindow(mw), sd(side)->varicongc,
		       sx, sy, mw->map.hw, mw->map.hh);
    } else {
	XSetClipMask(sdd(side), sd(side)->varicongc, None);
	XSetForeground(sdd(side), sd(side)->varicongc, side->ui->bgcolor);
	XDrawRectangle(sdd(side), XtWindow(mw), sd(side)->varicongc,
		       sx, sy, mw->map.hw, mw->map.hh);
	XSetForeground(sdd(side), sd(side)->varicongc, side->ui->fgcolor);
	XDrawRectangle(sdd(side), XtWindow(mw), sd(side)->varicongc,
		       sx, sy, mw->map.hw - 1, mw->map.hh - 1);
    }
#endif
}

#ifdef LINEAR

#endif /* LINEAR */

/* Draw an emblem identifying the given side.  If a side does not have a
   distinguishing emblem, fall back on some defaults. */

/* This needs to do something with masks or backgrounds or something... */

draw_side_emblem(side, mw, sx, sy, side2)
Side *side, *side2;
MapWidget mw;
int sx, sy;
{
    int p = mw->map.power, s2 = side_number(side2), ew, eh;
    long fg, bg;
    Pixmap epic = None;

    /* Decide whether the emblem ought to be drawn. (usually yes) */
    /* (what if we just want to see our own emblem?) */
    if (1 /* mustdraw
	|| (side != side2
	    && (numsides > 2 || side->ui->monochrome)) */) {
#if 0
	/* Decide on the colors to use with the emblem. */
	if (side->ui->numcolors[s2] > 0) {
	    fgcolor = side->ui->colors[s2][0];
	} else {
	    fgcolor = side->ui->fgcolor;
	}
	if (side->ui->numcolors[s2] > 1) {
	    bgcolor = side->ui->colors[s2][1];
	} else {
	    bgcolor = side->ui->bgcolor;
	}
#endif
	fg = mw->map.fgcolor;  bg = mw->core.background_pixel;
#if 0
	/* Decide what size of blob or picture to draw. */
	epic = sd(side)->emblempics[s2];
#endif
	if (epic == None) {
	    /* If no emblem, will make a colored blob. */
	    ew = max(1, uws[p] / 3);  eh = max(1, uhs[p] / 3);
	} else {
#if 0
	    ew = sd(side)->embw[s2];  eh = sd(side)->embh[s2];
	    /* If emblem won't fit in one quarter of unit cell at the
	       given power, then it will have to be a blob. */
	    if (ew > uws[power] / 2 || eh > uhs[power] / 2) {
		ew = max(1, uws[power] / 2);  eh = max(1, uhs[power] / 2);
		epic = None;
	    }
#endif
	}
	/* Draw a mask or background for the emblem. */
	if (0 /* mask defined */) {
	    /* Assume mask guarantees definition and contrast. */
	    XSetForeground(XtDisplay(mw), mw->map.gc, bg);
	    XSetClipMask(XtDisplay(mw), mw->map.gc, None);
	    XFillRectangle(XtDisplay(mw), XtWindow(mw), mw->map.gc,
			   sx, sy, ew, eh);
	} else if (0 /* low_contrast(side, fgcolor, colorunder) */) {
#if 0
	    /* Without a mask, we need bgcolor for contrast. */
	    bg = (0 /* low_contrast(side, fgcolor, side->ui->bgcolor) */ ?
		       side->ui->fgcolor : side->ui->bgcolor);
	    XSetForeground(XtDisplay(mw), mw->map.gc, bgcolor);
	    XSetClipMask(XtDisplay(mw), mw->map.gc, None);
	    /* Not good to expand by two pixels... */
	    XFillRectangle(XtDisplay(mw), XtWindow(mw), mw->map.gc,
			   sx-1, sy-1, ew+2, eh+2);
#endif
	} else {
	    /* Emblem alone should be distinctive enough. */
	}
	/* Now draw the emblem proper.  Note that we must *always* draw
	   it in its declared color. */
	XSetForeground(XtDisplay(mw), mw->map.gc, fg);
	XSetClipMask(XtDisplay(mw), mw->map.gc, epic);
	/* Don't need this if we're just doing blobs. */
	if (epic != None) {
	    XSetClipOrigin(XtDisplay(mw), mw->map.gc, sx, sy);
	}
	XFillRectangle(XtDisplay(mw), XtWindow(mw), mw->map.gc,
		       sx, sy, ew, eh);
    }
}

/* Draw something representing the population in the given hex.  Their side
   emblem alone is a good approximation. */

draw_people_emblem(side, win, sx, sy, power, side2)
Side *side, *side2;
Window win;
int power, sx, sy;
{
#if 0
    int s2 = side_number(side2);

    if (side2) {
	/* Shift the emblem to the middle of the hex. */
	/* (assumes emblem is defined!) */
	sx += hws[power] / 2 - sd(side)->embw[s2] / 2;
	sy += hhs[power] / 2 - sd(side)->embh[s2] / 2;
	draw_side_emblem(side, win, sx, sy, power, side2, TRUE,
			 side->ui->bgcolor /* where to get right val? */);
    } else {
	/* what to do for neutral peoples? use their emblem? */
    }
#endif
}

/* Splat up an appropriate icon. */

draw_blast_icon(side, win, sx, sy, type, color)
Side *side;
Window win;
int sx, sy, color;
char type;
{
#if 0
    XSetForeground(XtDisplay(mw), mw->map.gc, color);
    if (0 /* ever worthwhile to use font? */) {
	char buf[1];

	buf[0] = (type == 0 ? 'b' : (type == 1 ? 'c' : 'd'));
	sy += sd(side)->iconfont->max_bounds.ascent;
	XDrawString(XtDisplay(mw), win, mw->map.gc, sx, sy, buf, 1);
    } else {
	XSetClipMask(XtDisplay(mw), mw->map.gc, sd(side)->hitpics[type]);
	XSetClipOrigin(XtDisplay(mw), mw->map.gc, sx, sy);
	XFillRectangle(XtDisplay(mw), win, mw->map.gc, sx, sy, 20, 22);
    }
#endif
}

/* Flash the player's screen in an unmistakable way. */

invert_whole_map(side)
Side *side;
{
#if 0
    int sw = side->ui->vw * side->ui->hw, sh = side->ui->vh * side->ui->hh;

    XFillRectangle(XtDisplay(mw), side->ui->front->viewwin, sd(side)->flashgc, 0, 0, sw, sh);
    flush_output(side);
#endif
}

/* Draw just one of the mushroom cloud shapes. */

draw_mushroom(side, map, x, y, i)
Side *side;
Map *map;
int x, y, i;
{
#if 0
    int sx, sy;
    int color;

    color = ((side->ui->monochrome || i == 3) ? side->ui->fgcolor : side->ui->bgcolor);
    xform(side, map, x, y, &sx, &sy);
    XSetForeground(XtDisplay(mw), sd(side)->unitgc, color);
    XSetClipMask(XtDisplay(mw), sd(side)->unitgc, sd(side)->bombpics[i]);
    XSetClipOrigin(XtDisplay(mw), sd(side)->unitgc, sx-BH/4, sy-BH/2);
    XFillRectangle(XtDisplay(mw), side->ui->front->viewwin, sd(side)->unitgc, sx-BH/4, sy-BH/2, BH, BH);
    flush_output(side);
#endif
}

/* Return the actual dimensions of the unit icon that will be used at the
   given magnification. */

unit_icon_width(side, power, u)
Side *side;
int power, u;
{
    return sd(side)->unitw[power][u];
}

unit_icon_height(side, power, u)
Side *side;
int power, u;
{
    return sd(side)->unith[power][u];
}

/* Also draw a big 'X' through the losing side's maps. */

/* (This should draw fatter lines and invert rather than overwrite.) */

draw_map_scratchout(side, mw)
Side *side;
MapWidget mw;
{
#if 0
    XDrawLine(XtDisplay(mw), XtWindow(mw), sd(side)->textgc,
	      0, 0, mw->map.pxw, mw->map.pxh);
    XDrawLine(XtDisplay(mw), XtWindow(mw), sd(side)->textgc,
	      mw->map.pxw, 0, 0, mw->map.pxh);
#endif
}
