/* Copyright (c) 1987, 1988, 1991  Stanley T. Shebs. */
/* Copyright 1988 by Chris D. Peterson, MIT. */
/* Many improvements by Tim Moore, University of Utah. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact.  */

/* Interface implementations for the X11 version of xconq. */

#include "xtconf.h"

/* careful of the path of the X header files. */

#ifdef UNIX
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif UNIX

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Shell.h>

#include "xtmap.h"

#define BITMAPLIB "../pics"

/* How many magnifications we're supporting. */

#define NUMPOWERS 8

extern int mags[], hws[], hhs[], hcs[], uws[], uhs[];
extern int dolin[], bsx[8][7], bsy[8][7], lsx[8][6], lsy[8][6];


enum whattouse {
  dontdraw,
  useblocks,
  usepictures,
  usefontchars,
  usepolygons
};

/* Tools govern the interpretation of mouse clicks in a map view. */

enum tooltypes {
    LOOKTOOL = 0,
    UNITMOVETOOL = 1,
    UNITSHOOTTOOL = 2,
    HEXPAINTTOOL = 3,
    BORDPAINTTOOL = 4,
    CONNPAINTTOOL = 5,
    UNITADDTOOL = 6,
    NUMTOOLS = 7
};

#define MODENAMES { "Look", "Move", "Shoot", "Paint Hex", "Border", "Conn" }

#define INFOLINES 4

/* Bounding box size for bomb images. */

#define BH 32

/* The user interface substructure.  This is only allocated for sides with
   attached displays. */

/* Some of the slots may be casted to device-specific slots - their type
   should be big enough for large integers and pointers both. */

typedef struct a_ui {
    short followaction;       /* move to where a change has occured */
    /* input hacking */
    struct a_map *inpmap;     /* which map all this was on */
    struct a_map *downmap;    /* which map all this was on */
    int downx, downy, downdir;   /* press map coordinates and hexside */
    struct a_map *upmap;      /* which map all this was on */
    int upx, upy, updir;      /* release map coordinates and hexside */

    /* Random input data slots - flush these! */
    char reqch;               /* keyboard char */
    struct a_side *reqoside;
    struct a_unit *requnit;
    int reqvalue, reqvalue2;
    int reqstrbeg, reqcurstr; /* data about string under construction */
    char *reqdeflt;           /* A default string */
    int reqposx, reqposy;     /* accumulator for position reading */
    char ustr[MAXUTYPES];     /* used in composing unit type hints */
    int uvec[MAXUTYPES];      /* vector of allowed unit types to input */
    int bvec[MAXUTYPES];      /* bit vector of allowed unit types to input */
    /* Machinery for standing orders */
    int teach;               /* true when only setting a standing order */
    struct a_unit *sounit;    /* unit that stores standing order for occs */
    int soutype;              /* unit type that will get standing orders */
    struct a_order *tmporder; /* holding place for orders */


    /* Constructed during display init */
    short nw, nh;             /* length and number of notice lines */
    short fw, fh;             /* dimensions of text font (in pixels) */
    short margin;         /* spacing around things */
    /* Working variables for the display */
    short lastx, lasty;       /* last current x and y (-1,-1 initially) */
    char *noticebuf[MAXNOTES];/* data for the notice area */
    char promptbuf[BUFSIZE];  /* current prompt + input str on display */
    short bottom_note;        /* bottom note being displayed. */
    short redisplay_ok;       /* Process redisplay events */
    /* All the colors used - filled in by display init */
    long bdcolor;             /* color for borders */
    long bgcolor;             /* background color */
    long fgcolor;             /* foreground (text) color */
    long owncolor;            /* color for us (usually black) */
    long altcolor;            /* another color for us (usually blue) */
    long enemycolor;          /* color for them (usually red) */
    long neutcolor;           /* color for fencesitters (usually gray) */
    long graycolor;           /* color for graying out (usually gray) */
    long diffcolor;           /* unusual/distinct color (usually maroon) */
    long goodcolor;           /* color for OKness (usually green) */
    long badcolor;            /* color for non-OKness (usually red) */
    long hexfgcolor[MAXTTYPES]; /* the "fg" color of each terrain type */
    long ourcolors[3];        /* national colors */
    long ournumcolors;
    long ouremblem;          /* pattern for national emblem (pixmap?) */
    long emblems[MAXSIDES];
    long colors[MAXSIDES][3];
    int numcolors[MAXSIDES];
    Widget game;                /* overall game info */
    Widget msg;                 /* places for notices/warnings */
    Widget clock;               /* chess clock display */
    struct a_map *maps;       /* chain of maps that we're using */
    struct a_map *front;      /* the one that's out in front */
    struct a_map *world;      /* the one that shows everything */
    struct a_list *lists;     /* chain of lists */
    struct a_list *frontlist;
    struct a_closeup *closeups;     /* chain of unit closeups */
    struct a_closeup *frontcloseup;
    struct a_side_closeup *sidecloseups;     /* chain of side closeups */
    struct a_side_closeup *ourcloseup;
    Widget help;                /* help window */
    Widget helpinfo;            /* help content subwindow */
    struct a_helpnode *curhelpnode;
    int active;             /* side's specific display structure */
    Display *dpy;
    Widget shell;
    GC gc;			/* a tmp graphics context for this display */
    GC textgc;			/* foreground on background text */
    GC icongc;			/* icon graphics context */
    GC invicongc;		/* icon gc inverted colors. */
    GC varicongc;		/* This is an icon gc with variable fgcolor. */
    GC unitgc;			/* unit icon display gc. */
    GC cleargc;			/* The gc to use for clearing areas. */
    GC clearbitgc;              /* gc for clearing bitmaps */
    XFontStruct *textfont;	/* font for text display */
    /* need legend fonts for other things too */
    XFontStruct *cursorfont;      /* font for cursors */
    /* Terrain drawing machinery */
    enum whattouse usewhat[NUMPOWERS][MAXTTYPES];
    Pixmap terrpics[NUMPOWERS][MAXTTYPES]; /* terrain bitmaps */
    char terrchars[NUMPOWERS][MAXTTYPES];
    XFontStruct *terrfonts[NUMPOWERS][MAXTTYPES];
    Pixmap hexpics[NUMPOWERS], bhexpics[NUMPOWERS];
    Pixmap hexchars[NUMPOWERS], bhexchars[NUMPOWERS];
    XFontStruct *hexfonts[NUMPOWERS], *bhexfonts[NUMPOWERS];
    XFontStruct *iconfont;
    /* Unit drawing machinery */
    Pixmap unitpics[NUMPOWERS][MAXUTYPES];	/* unit bitmaps */
    Pixmap unitmasks[NUMPOWERS][MAXUTYPES];
    char unitchars[NUMPOWERS][MAXUTYPES];
    XFontStruct *unitfonts[NUMPOWERS][MAXUTYPES];
    XFontStruct *unitfont;	/* font for unit characters */
    int unitw[NUMPOWERS][MAXUTYPES], unith[NUMPOWERS][MAXUTYPES];
    /* Emblem drawing machinery */
    Pixmap emblempics[MAXSIDES];
    Pixmap emblemmasks[MAXSIDES];
    int embw[MAXSIDES], embh[MAXSIDES];
    /* Map legend drawing machinery */
    XFontStruct *ulegendfonts[NUMPOWERS][MAXUTYPES];
    XFontStruct *flegendfonts[NUMPOWERS][10];  /* num feature types? */
    /* Random stuff */
    Pixmap bombpics[4];		/* mushroom clouds */
    Pixmap hitpics[3];
    Pixmap boxcurs, boxmask;

    Widget datelabel;
    Widget sidenamelist;
    char *sidestrings[MAXSIDES];

    Widget sides;

    Widget yesnodialog;
    Widget utypedialog;
    Widget ttypedialog;
    Widget cmdmenu;
} UI;

/* Each side can open up any number and shape of maps. */

/* All the maps will share colors and bitmaps though. */

typedef struct a_map {
    int isworld;
    short curx, cury;         /* current spot being looked at */
    struct a_unit *curunit;   /* unit under cursor */
    short pxw, pxh;           /* size of window in pixels */
    short vx, vy;             /* lower left of viewport in hex coords */
    short vw, vh;             /* Boundaries of viewport in hex coords */
    short vw2, vh2;           /* 1/2 (rounded down) of above values */
    short power;              /* index to this map's magnification */
    short mag;                /* magnification of each hex (a power of 2) */
    short vcx, vcy;           /* center of the view */
    short lastvcx, lastvcy;
    short hw, hh;             /* pixel dims of a hex cell */
    short hch;                /* height of hex cell between centers */
    short uw, uh;             /* pixel dims of unit subcell */
    Widget mapwin;
    Widget viewwin;
    /* How to draw the map. */
    int drawterrain;         /* Display solid color terrain on the map? */
    int drawhexedges;        /* Draw outlines around hexes? */
    int drawhexpats;         /* Draw terrain patterns? */
    int drawunits;           /* Display units on the map? */
    int drawnames;           /* Display unit names/numbers on the map? */
    short oldesttoshow;       /* the relative time of the oldest data */
    short agetofade;          /* point at which to gray out the display */
    short newesttoshow;       /* the relative time of the newest data */
    short fadeterrain;        /* fade world features as well as units? */
#ifdef LEVELS
    short lowestlevel;        /* the lowest level of units to display */
    short highestlevel;       /* the highest level of units to display */
#endif /* LEVELS */
#ifdef LINEAR
    int drawborders;
    int drawconns;
#endif /* LINEAR */
    int drawpeople;
    int drawelevations;
    int drawresources[MAXMTYPES];
    short drawtemp;
    short drawweather;
    /* The state of the control panel. */
    short curtool;       /* interpretation of mouse clicks */
    short curshifttool;  /* interp of shifted clicks */

    short curutype;
    short curttype;
    short curbordtype;
    short curconntype;
    short curside;
    short curfeature;
    /* Link to the next map. */
    struct a_map *next;
} Map;

/* Each side can open up any number of lists of units. */

typedef struct a_list {
    int type;               /* what this list is displaying. */
    int subtype;
    int test;
    int value;
    int sorton;
    int sortorder;
    int listlen;
    char **listbuf;
    Widget listwin;
    Widget listwininfo;
    struct a_list *next;
} List;

/* Each side can open up any number of lists of units. */

typedef struct a_closeup {
    struct a_unit *unit;
    Widget closeupwin;
    struct a_closeup *next;
} Closeup;

/* A closer look at a side. */

typedef struct a_side_closeup {
    struct a_side *side;
    Widget closeupwin;
    struct a_side_closeup *next;
} SideCloseup;

/* Iteration over all of a side's maps. */

#define for_all_maps(s,m)  \
  for ((m) = (s)->ui->maps; (m) != NULL; (m) = (m)->next)

/* Iteration over all of a side's lists. */

#define for_all_lists(s,l)  \
  for ((l) = (s)->ui->lists; (l) != NULL; (l) = (l)->next)

/* Iteration over all of a side's unit closeups. */

#define for_all_closeups(s,c)  \
  for ((c) = (s)->ui->closeups; (c) != NULL; (c) = (c)->next)

/* Iteration over all of a side's side closeups. */

#define for_all_side_closeups(s,c)  \
  for ((c) = (s)->ui->sidecloseups; (c) != NULL; (c) = (c)->next)

/* Casted etc accessors to X structures. */

#define sd(s) ((s)->ui)

#define sdd(s) ((s)->ui->dpy)

#define toplevel(s) ((s)->ui->shell)

/* Random function declarations. */

XFontStruct * open_font();
Pixmap load_bitmap();
Cursor make_cursor();

/* Values of parameters generally tied to fonts and the like. */


extern int margin, bd;

extern int conwid, conhgt;
extern int sbarwid;

extern int rootcursor, windows_up;

extern Widget thistoplevel;
extern Widget setup;

extern Arg tmpargs[];

extern XtAppContext thisapp;

/* Magic values (unlikely to show up as player input) */

#define DFLTFLAG (-12345)       /* flags when default arg to be used */
#define ITERFLAG (-12347)       /* flags args that are "itertime" */
#define NEXTPAGE (-101010)      /* Print these commands on the next page */

#define ESCAPE '\033'           /* standard abort character */
#define BACKSPACE '\010'        /* for fixing strings being entered */

Unit *find_unit_in_list();
Side *side_from_display();
